#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_wireless_controller_wids_profile
short_description: Configure wireless intrusion detection system (WIDS) profiles in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS (FOS) device by allowing the
      user to set and modify wireless_controller feature and wids_profile category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.5
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
        description:
            - FortiOS or FortiGate IP address.
        type: str
        required: false
    username:
        description:
            - FortiOS or FortiGate username.
        type: str
        required: false
    password:
        description:
            - FortiOS or FortiGate password.
        type: str
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        type: str
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS protocol.
        type: bool
        default: true
    ssl_verify:
        description:
            - Ensures FortiGate certificate must be verified by a proper CA.
        type: bool
        default: true
        version_added: 2.9
    state:
        description:
            - Indicates whether to create or remove the object.
              This attribute was present already in previous version in a deeper level.
              It has been moved out to this outer level.
        type: str
        required: false
        choices:
            - present
            - absent
        version_added: 2.9
    wireless_controller_wids_profile:
        description:
            - Configure wireless intrusion detection system (WIDS) profiles.
        default: null
        type: dict
        suboptions:
            state:
                description:
                    - B(Deprecated)
                    - Starting with Ansible 2.9 we recommend using the top-level 'state' parameter.
                    - HORIZONTALLINE
                    - Indicates whether to create or remove the object.
                type: str
                required: false
                choices:
                    - present
                    - absent
            ap_auto_suppress:
                description:
                    - Enable/disable on-wire rogue AP auto-suppression .
                type: str
                choices:
                    - enable
                    - disable
            ap_bgscan_disable_day:
                description:
                    - Optionally turn off scanning for one or more days of the week. Separate the days with a space. By default, no days are set.
                type: str
                choices:
                    - sunday
                    - monday
                    - tuesday
                    - wednesday
                    - thursday
                    - friday
                    - saturday
            ap_bgscan_disable_end:
                description:
                    - "End time, using a 24-hour clock in the format of hh:mm, for disabling background scanning ."
                type: str
            ap_bgscan_disable_start:
                description:
                    - "Start time, using a 24-hour clock in the format of hh:mm, for disabling background scanning ."
                type: str
            ap_bgscan_duration:
                description:
                    - Listening time on a scanning channel (10 - 1000 msec).
                type: int
            ap_bgscan_idle:
                description:
                    - Waiting time for channel inactivity before scanning this channel (0 - 1000 msec).
                type: int
            ap_bgscan_intv:
                description:
                    - Period of time between scanning two channels (1 - 600 sec).
                type: int
            ap_bgscan_period:
                description:
                    - Period of time between background scans (60 - 3600 sec).
                type: int
            ap_bgscan_report_intv:
                description:
                    - Period of time between background scan reports (15 - 600 sec).
                type: int
            ap_fgscan_report_intv:
                description:
                    - Period of time between foreground scan reports (15 - 600 sec).
                type: int
            ap_scan:
                description:
                    - Enable/disable rogue AP detection.
                type: str
                choices:
                    - disable
                    - enable
            ap_scan_passive:
                description:
                    - Enable/disable passive scanning. Enable means do not send probe request on any channels .
                type: str
                choices:
                    - enable
                    - disable
            asleap_attack:
                description:
                    - Enable/disable asleap attack detection .
                type: str
                choices:
                    - enable
                    - disable
            assoc_flood_thresh:
                description:
                    - The threshold value for association frame flooding.
                type: int
            assoc_flood_time:
                description:
                    - Number of seconds after which a station is considered not connected.
                type: int
            assoc_frame_flood:
                description:
                    - Enable/disable association frame flooding detection .
                type: str
                choices:
                    - enable
                    - disable
            auth_flood_thresh:
                description:
                    - The threshold value for authentication frame flooding.
                type: int
            auth_flood_time:
                description:
                    - Number of seconds after which a station is considered not connected.
                type: int
            auth_frame_flood:
                description:
                    - Enable/disable authentication frame flooding detection .
                type: str
                choices:
                    - enable
                    - disable
            comment:
                description:
                    - Comment.
                type: str
            deauth_broadcast:
                description:
                    - Enable/disable broadcasting de-authentication detection .
                type: str
                choices:
                    - enable
                    - disable
            deauth_unknown_src_thresh:
                description:
                    - "Threshold value per second to deauth unknown src for DoS attack (0: no limit)."
                type: int
            eapol_fail_flood:
                description:
                    - Enable/disable EAPOL-Failure flooding (to AP) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_fail_intv:
                description:
                    - The detection interval for EAPOL-Failure flooding (1 - 3600 sec).
                type: int
            eapol_fail_thresh:
                description:
                    - The threshold value for EAPOL-Failure flooding in specified interval.
                type: int
            eapol_logoff_flood:
                description:
                    - Enable/disable EAPOL-Logoff flooding (to AP) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_logoff_intv:
                description:
                    - The detection interval for EAPOL-Logoff flooding (1 - 3600 sec).
                type: int
            eapol_logoff_thresh:
                description:
                    - The threshold value for EAPOL-Logoff flooding in specified interval.
                type: int
            eapol_pre_fail_flood:
                description:
                    - Enable/disable premature EAPOL-Failure flooding (to STA) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_pre_fail_intv:
                description:
                    - The detection interval for premature EAPOL-Failure flooding (1 - 3600 sec).
                type: int
            eapol_pre_fail_thresh:
                description:
                    - The threshold value for premature EAPOL-Failure flooding in specified interval.
                type: int
            eapol_pre_succ_flood:
                description:
                    - Enable/disable premature EAPOL-Success flooding (to STA) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_pre_succ_intv:
                description:
                    - The detection interval for premature EAPOL-Success flooding (1 - 3600 sec).
                type: int
            eapol_pre_succ_thresh:
                description:
                    - The threshold value for premature EAPOL-Success flooding in specified interval.
                type: int
            eapol_start_flood:
                description:
                    - Enable/disable EAPOL-Start flooding (to AP) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_start_intv:
                description:
                    - The detection interval for EAPOL-Start flooding (1 - 3600 sec).
                type: int
            eapol_start_thresh:
                description:
                    - The threshold value for EAPOL-Start flooding in specified interval.
                type: int
            eapol_succ_flood:
                description:
                    - Enable/disable EAPOL-Success flooding (to AP) detection .
                type: str
                choices:
                    - enable
                    - disable
            eapol_succ_intv:
                description:
                    - The detection interval for EAPOL-Success flooding (1 - 3600 sec).
                type: int
            eapol_succ_thresh:
                description:
                    - The threshold value for EAPOL-Success flooding in specified interval.
                type: int
            invalid_mac_oui:
                description:
                    - Enable/disable invalid MAC OUI detection.
                type: str
                choices:
                    - enable
                    - disable
            long_duration_attack:
                description:
                    - Enable/disable long duration attack detection based on user configured threshold .
                type: str
                choices:
                    - enable
                    - disable
            long_duration_thresh:
                description:
                    - Threshold value for long duration attack detection (1000 - 32767 usec).
                type: int
            name:
                description:
                    - WIDS profile name.
                required: true
                type: str
            null_ssid_probe_resp:
                description:
                    - Enable/disable null SSID probe response detection .
                type: str
                choices:
                    - enable
                    - disable
            sensor_mode:
                description:
                    - Scan WiFi nearby stations .
                type: str
                choices:
                    - disable
                    - foreign
                    - both
            spoofed_deauth:
                description:
                    - Enable/disable spoofed de-authentication attack detection .
                type: str
                choices:
                    - enable
                    - disable
            weak_wep_iv:
                description:
                    - Enable/disable weak WEP IV (Initialization Vector) detection .
                type: str
                choices:
                    - enable
                    - disable
            wireless_bridge:
                description:
                    - Enable/disable wireless bridge detection .
                type: str
                choices:
                    - enable
                    - disable
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
   ssl_verify: "False"
  tasks:
  - name: Configure wireless intrusion detection system (WIDS) profiles.
    fortios_wireless_controller_wids_profile:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      state: "present"
      wireless_controller_wids_profile:
        ap_auto_suppress: "enable"
        ap_bgscan_disable_day: "sunday"
        ap_bgscan_disable_end: "<your_own_value>"
        ap_bgscan_disable_start: "<your_own_value>"
        ap_bgscan_duration: "7"
        ap_bgscan_idle: "8"
        ap_bgscan_intv: "9"
        ap_bgscan_period: "10"
        ap_bgscan_report_intv: "11"
        ap_fgscan_report_intv: "12"
        ap_scan: "disable"
        ap_scan_passive: "enable"
        asleap_attack: "enable"
        assoc_flood_thresh: "16"
        assoc_flood_time: "17"
        assoc_frame_flood: "enable"
        auth_flood_thresh: "19"
        auth_flood_time: "20"
        auth_frame_flood: "enable"
        comment: "Comment."
        deauth_broadcast: "enable"
        deauth_unknown_src_thresh: "24"
        eapol_fail_flood: "enable"
        eapol_fail_intv: "26"
        eapol_fail_thresh: "27"
        eapol_logoff_flood: "enable"
        eapol_logoff_intv: "29"
        eapol_logoff_thresh: "30"
        eapol_pre_fail_flood: "enable"
        eapol_pre_fail_intv: "32"
        eapol_pre_fail_thresh: "33"
        eapol_pre_succ_flood: "enable"
        eapol_pre_succ_intv: "35"
        eapol_pre_succ_thresh: "36"
        eapol_start_flood: "enable"
        eapol_start_intv: "38"
        eapol_start_thresh: "39"
        eapol_succ_flood: "enable"
        eapol_succ_intv: "41"
        eapol_succ_thresh: "42"
        invalid_mac_oui: "enable"
        long_duration_attack: "enable"
        long_duration_thresh: "45"
        name: "default_name_46"
        null_ssid_probe_resp: "enable"
        sensor_mode: "disable"
        spoofed_deauth: "enable"
        weak_wep_iv: "enable"
        wireless_bridge: "enable"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import Connection
from ansible.module_utils.network.fortios.fortios import FortiOSHandler
from ansible.module_utils.network.fortimanager.common import FAIL_SOCKET_MSG


def login(data, fos):
    host = data['host']
    username = data['username']
    password = data['password']
    ssl_verify = data['ssl_verify']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password, verify=ssl_verify)


def filter_wireless_controller_wids_profile_data(json):
    option_list = ['ap_auto_suppress', 'ap_bgscan_disable_day', 'ap_bgscan_disable_end',
                   'ap_bgscan_disable_start', 'ap_bgscan_duration', 'ap_bgscan_idle',
                   'ap_bgscan_intv', 'ap_bgscan_period', 'ap_bgscan_report_intv',
                   'ap_fgscan_report_intv', 'ap_scan', 'ap_scan_passive',
                   'asleap_attack', 'assoc_flood_thresh', 'assoc_flood_time',
                   'assoc_frame_flood', 'auth_flood_thresh', 'auth_flood_time',
                   'auth_frame_flood', 'comment', 'deauth_broadcast',
                   'deauth_unknown_src_thresh', 'eapol_fail_flood', 'eapol_fail_intv',
                   'eapol_fail_thresh', 'eapol_logoff_flood', 'eapol_logoff_intv',
                   'eapol_logoff_thresh', 'eapol_pre_fail_flood', 'eapol_pre_fail_intv',
                   'eapol_pre_fail_thresh', 'eapol_pre_succ_flood', 'eapol_pre_succ_intv',
                   'eapol_pre_succ_thresh', 'eapol_start_flood', 'eapol_start_intv',
                   'eapol_start_thresh', 'eapol_succ_flood', 'eapol_succ_intv',
                   'eapol_succ_thresh', 'invalid_mac_oui', 'long_duration_attack',
                   'long_duration_thresh', 'name', 'null_ssid_probe_resp',
                   'sensor_mode', 'spoofed_deauth', 'weak_wep_iv',
                   'wireless_bridge']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def underscore_to_hyphen(data):
    if isinstance(data, list):
        for elem in data:
            elem = underscore_to_hyphen(elem)
    elif isinstance(data, dict):
        new_data = {}
        for k, v in data.items():
            new_data[k.replace('_', '-')] = underscore_to_hyphen(v)
        data = new_data

    return data


def wireless_controller_wids_profile(data, fos):
    vdom = data['vdom']
    if 'state' in data and data['state']:
        state = data['state']
    elif 'state' in data['wireless_controller_wids_profile'] and data['wireless_controller_wids_profile']:
        state = data['wireless_controller_wids_profile']['state']
    else:
        state = True
    wireless_controller_wids_profile_data = data['wireless_controller_wids_profile']
    filtered_data = underscore_to_hyphen(filter_wireless_controller_wids_profile_data(wireless_controller_wids_profile_data))

    if state == "present":
        return fos.set('wireless-controller',
                       'wids-profile',
                       data=filtered_data,
                       vdom=vdom)

    elif state == "absent":
        return fos.delete('wireless-controller',
                          'wids-profile',
                          mkey=filtered_data['name'],
                          vdom=vdom)


def is_successful_status(status):
    return status['status'] == "success" or \
        status['http_method'] == "DELETE" and status['http_status'] == 404


def fortios_wireless_controller(data, fos):

    if data['wireless_controller_wids_profile']:
        resp = wireless_controller_wids_profile(data, fos)

    return not is_successful_status(resp), \
        resp['status'] == "success", \
        resp


def main():
    fields = {
        "host": {"required": False, "type": "str"},
        "username": {"required": False, "type": "str"},
        "password": {"required": False, "type": "str", "default": "", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "ssl_verify": {"required": False, "type": "bool", "default": True},
        "state": {"required": False, "type": "str",
                  "choices": ["present", "absent"]},
        "wireless_controller_wids_profile": {
            "required": False, "type": "dict", "default": None,
            "options": {
                "state": {"required": False, "type": "str",
                          "choices": ["present", "absent"]},
                "ap_auto_suppress": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "ap_bgscan_disable_day": {"required": False, "type": "str",
                                          "choices": ["sunday", "monday", "tuesday",
                                                      "wednesday", "thursday", "friday",
                                                      "saturday"]},
                "ap_bgscan_disable_end": {"required": False, "type": "str"},
                "ap_bgscan_disable_start": {"required": False, "type": "str"},
                "ap_bgscan_duration": {"required": False, "type": "int"},
                "ap_bgscan_idle": {"required": False, "type": "int"},
                "ap_bgscan_intv": {"required": False, "type": "int"},
                "ap_bgscan_period": {"required": False, "type": "int"},
                "ap_bgscan_report_intv": {"required": False, "type": "int"},
                "ap_fgscan_report_intv": {"required": False, "type": "int"},
                "ap_scan": {"required": False, "type": "str",
                            "choices": ["disable", "enable"]},
                "ap_scan_passive": {"required": False, "type": "str",
                                    "choices": ["enable", "disable"]},
                "asleap_attack": {"required": False, "type": "str",
                                  "choices": ["enable", "disable"]},
                "assoc_flood_thresh": {"required": False, "type": "int"},
                "assoc_flood_time": {"required": False, "type": "int"},
                "assoc_frame_flood": {"required": False, "type": "str",
                                      "choices": ["enable", "disable"]},
                "auth_flood_thresh": {"required": False, "type": "int"},
                "auth_flood_time": {"required": False, "type": "int"},
                "auth_frame_flood": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "comment": {"required": False, "type": "str"},
                "deauth_broadcast": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "deauth_unknown_src_thresh": {"required": False, "type": "int"},
                "eapol_fail_flood": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "eapol_fail_intv": {"required": False, "type": "int"},
                "eapol_fail_thresh": {"required": False, "type": "int"},
                "eapol_logoff_flood": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                "eapol_logoff_intv": {"required": False, "type": "int"},
                "eapol_logoff_thresh": {"required": False, "type": "int"},
                "eapol_pre_fail_flood": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "eapol_pre_fail_intv": {"required": False, "type": "int"},
                "eapol_pre_fail_thresh": {"required": False, "type": "int"},
                "eapol_pre_succ_flood": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "eapol_pre_succ_intv": {"required": False, "type": "int"},
                "eapol_pre_succ_thresh": {"required": False, "type": "int"},
                "eapol_start_flood": {"required": False, "type": "str",
                                      "choices": ["enable", "disable"]},
                "eapol_start_intv": {"required": False, "type": "int"},
                "eapol_start_thresh": {"required": False, "type": "int"},
                "eapol_succ_flood": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "eapol_succ_intv": {"required": False, "type": "int"},
                "eapol_succ_thresh": {"required": False, "type": "int"},
                "invalid_mac_oui": {"required": False, "type": "str",
                                    "choices": ["enable", "disable"]},
                "long_duration_attack": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "long_duration_thresh": {"required": False, "type": "int"},
                "name": {"required": True, "type": "str"},
                "null_ssid_probe_resp": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "sensor_mode": {"required": False, "type": "str",
                                "choices": ["disable", "foreign", "both"]},
                "spoofed_deauth": {"required": False, "type": "str",
                                   "choices": ["enable", "disable"]},
                "weak_wep_iv": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "wireless_bridge": {"required": False, "type": "str",
                                    "choices": ["enable", "disable"]}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)

    # legacy_mode refers to using fortiosapi instead of HTTPAPI
    legacy_mode = 'host' in module.params and module.params['host'] is not None and \
                  'username' in module.params and module.params['username'] is not None and \
                  'password' in module.params and module.params['password'] is not None

    if not legacy_mode:
        if module._socket_path:
            connection = Connection(module._socket_path)
            fos = FortiOSHandler(connection)

            is_error, has_changed, result = fortios_wireless_controller(module.params, fos)
        else:
            module.fail_json(**FAIL_SOCKET_MSG)
    else:
        try:
            from fortiosapi import FortiOSAPI
        except ImportError:
            module.fail_json(msg="fortiosapi module is required")

        fos = FortiOSAPI()

        login(module.params, fos)
        is_error, has_changed, result = fortios_wireless_controller(module.params, fos)
        fos.logout()

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function

__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1', 'status': ["preview"], 'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_compute_target_pool_info
description:
- Gather info for GCP TargetPool
- This module was called C(gcp_compute_target_pool_facts) before Ansible 2.9. The
  usage has not changed.
short_description: Gather info for GCP TargetPool
version_added: 2.7
author: Google Inc. (@googlecloudplatform)
requirements:
- python >= 2.6
- requests >= 2.18.4
- google-auth >= 1.3.0
options:
  filters:
    description:
    - A list of filter value pairs. Available filters are listed here U(https://cloud.google.com/sdk/gcloud/reference/topic/filters).
    - Each additional filter in the list will act be added as an AND condition (filter1
      and filter2) .
    type: list
  region:
    description:
    - The region where the target pool resides.
    required: true
    type: str
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: get info on a target pool
  gcp_compute_target_pool_info:
    region: us-west1
    filters:
    - name = test_object
    project: test_project
    auth_kind: serviceaccount
    service_account_file: "/tmp/auth.pem"
'''

RETURN = '''
resources:
  description: List of resources
  returned: always
  type: complex
  contains:
    backupPool:
      description:
      - This field is applicable only when the containing target pool is serving a
        forwarding rule as the primary pool, and its failoverRatio field is properly
        set to a value between [0, 1].
      - 'backupPool and failoverRatio together define the fallback behavior of the
        primary target pool: if the ratio of the healthy instances in the primary
        pool is at or below failoverRatio, traffic arriving at the load-balanced IP
        will be directed to the backup pool.'
      - In case where failoverRatio and backupPool are not set, or all the instances
        in the backup pool are unhealthy, the traffic will be directed back to the
        primary pool in the "force" mode, where traffic will be spread to the healthy
        instances with the best effort, or to all instances when no instance is healthy.
      returned: success
      type: dict
    creationTimestamp:
      description:
      - Creation timestamp in RFC3339 text format.
      returned: success
      type: str
    description:
      description:
      - An optional description of this resource.
      returned: success
      type: str
    failoverRatio:
      description:
      - This field is applicable only when the containing target pool is serving a
        forwarding rule as the primary pool (i.e., not as a backup pool to some other
        target pool). The value of the field must be in [0, 1].
      - 'If set, backupPool must also be set. They together define the fallback behavior
        of the primary target pool: if the ratio of the healthy instances in the primary
        pool is at or below this number, traffic arriving at the load-balanced IP
        will be directed to the backup pool.'
      - In case where failoverRatio is not set or all the instances in the backup
        pool are unhealthy, the traffic will be directed back to the primary pool
        in the "force" mode, where traffic will be spread to the healthy instances
        with the best effort, or to all instances when no instance is healthy.
      returned: success
      type: str
    healthCheck:
      description:
      - A reference to a HttpHealthCheck resource.
      - A member instance in this pool is considered healthy if and only if the health
        checks pass. If not specified it means all member instances will be considered
        healthy at all times.
      returned: success
      type: dict
    id:
      description:
      - The unique identifier for the resource.
      returned: success
      type: int
    instances:
      description:
      - A list of virtual machine instances serving this pool.
      - They must live in zones contained in the same region as this pool.
      returned: success
      type: list
    name:
      description:
      - Name of the resource. Provided by the client when the resource is created.
        The name must be 1-63 characters long, and comply with RFC1035. Specifically,
        the name must be 1-63 characters long and match the regular expression `[a-z]([-a-z0-9]*[a-z0-9])?`
        which means the first character must be a lowercase letter, and all following
        characters must be a dash, lowercase letter, or digit, except the last character,
        which cannot be a dash.
      returned: success
      type: str
    sessionAffinity:
      description:
      - 'Session affinity option. Must be one of these values: - NONE: Connections
        from the same client IP may go to any instance in the pool.'
      - "- CLIENT_IP: Connections from the same client IP will go to the same instance
        in the pool while that instance remains healthy."
      - "- CLIENT_IP_PROTO: Connections from the same client IP with the same IP protocol
        will go to the same instance in the pool while that instance remains healthy."
      returned: success
      type: str
    region:
      description:
      - The region where the target pool resides.
      returned: success
      type: str
'''

################################################################################
# Imports
################################################################################
from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest
import json

################################################################################
# Main
################################################################################


def main():
    module = GcpModule(argument_spec=dict(filters=dict(type='list', elements='str'), region=dict(required=True, type='str')))

    if module._name == 'gcp_compute_target_pool_facts':
        module.deprecate("The 'gcp_compute_target_pool_facts' module has been renamed to 'gcp_compute_target_pool_info'", version='2.13')

    if not module.params['scopes']:
        module.params['scopes'] = ['https://www.googleapis.com/auth/compute']

    return_value = {'resources': fetch_list(module, collection(module), query_options(module.params['filters']))}
    module.exit_json(**return_value)


def collection(module):
    return "https://www.googleapis.com/compute/v1/projects/{project}/regions/{region}/targetPools".format(**module.params)


def fetch_list(module, link, query):
    auth = GcpSession(module, 'compute')
    return auth.list(link, return_if_object, array_name='items', params={'filter': query})


def query_options(filters):
    if not filters:
        return ''

    if len(filters) == 1:
        return filters[0]
    else:
        queries = []
        for f in filters:
            # For multiple queries, all queries should have ()
            if f[0] != '(' and f[-1] != ')':
                queries.append("(%s)" % ''.join(f))
            else:
                queries.append(f)

        return ' '.join(queries)


def return_if_object(module, response):
    # If not found, return nothing.
    if response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    try:
        module.raise_for_status(response)
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError) as inst:
        module.fail_json(msg="Invalid JSON response with error: %s" % inst)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))

    return result


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Guido Günther <agx@sigxcpu.org>, Daniel Lobato Garcia <dlobatog@redhat.com>
# Copyright (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    name: foreman
    plugin_type: inventory
    short_description: foreman inventory source
    version_added: "2.6"
    requirements:
        - requests >= 1.1
    description:
        - Get inventory hosts from the foreman service.
        - "Uses a configuration file as an inventory source, it must end in ``.foreman.yml`` or ``.foreman.yaml`` and has a ``plugin: foreman`` entry."
    extends_documentation_fragment:
        - inventory_cache
    options:
      plugin:
        description: the name of this plugin, it should always be set to 'foreman' for this plugin to recognize it as it's own.
        required: True
        choices: ['foreman']
      url:
        description: url to foreman
        default: 'http://localhost:3000'
        env:
            - name: FOREMAN_SERVER
              version_added: "2.8"
      user:
        description: foreman authentication user
        required: True
        env:
            - name: FOREMAN_USER
              version_added: "2.8"
      password:
        description: foreman authentication password
        required: True
        env:
            - name: FOREMAN_PASSWORD
              version_added: "2.8"
      validate_certs:
        description: verify SSL certificate if using https
        type: boolean
        default: False
      group_prefix:
        description: prefix to apply to foreman groups
        default: foreman_
      vars_prefix:
        description: prefix to apply to host variables, does not include facts nor params
        default: foreman_
      want_facts:
        description: Toggle, if True the plugin will retrieve host facts from the server
        type: boolean
        default: False
      want_params:
        description: Toggle, if true the inventory will retrieve 'all_parameters' information as host vars
        type: boolean
        default: False
'''

EXAMPLES = '''
# my.foreman.yml
plugin: foreman
url: http://localhost:2222
user: ansible-tester
password: secure
validate_certs: False
'''

from distutils.version import LooseVersion

from ansible.errors import AnsibleError
from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.module_utils.common._collections_compat import MutableMapping
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, to_safe_group_name

# 3rd party imports
try:
    import requests
    if LooseVersion(requests.__version__) < LooseVersion('1.1.0'):
        raise ImportError
except ImportError:
    raise AnsibleError('This script requires python-requests 1.1 as a minimum version')

from requests.auth import HTTPBasicAuth


class InventoryModule(BaseInventoryPlugin, Cacheable):
    ''' Host inventory parser for ansible using foreman as source. '''

    NAME = 'foreman'

    def __init__(self):

        super(InventoryModule, self).__init__()

        # from config
        self.foreman_url = None

        self.session = None
        self.cache_key = None
        self.use_cache = None

    def verify_file(self, path):

        valid = False
        if super(InventoryModule, self).verify_file(path):
            if path.endswith(('foreman.yaml', 'foreman.yml')):
                valid = True
            else:
                self.display.vvv('Skipping due to inventory source not ending in "foreman.yaml" nor "foreman.yml"')
        return valid

    def _get_session(self):
        if not self.session:
            self.session = requests.session()
            self.session.auth = HTTPBasicAuth(self.get_option('user'), to_bytes(self.get_option('password')))
            self.session.verify = self.get_option('validate_certs')
        return self.session

    def _get_json(self, url, ignore_errors=None):

        if not self.use_cache or url not in self._cache.get(self.cache_key, {}):

            if self.cache_key not in self._cache:
                self._cache[self.cache_key] = {url: ''}

            results = []
            s = self._get_session()
            params = {'page': 1, 'per_page': 250}
            while True:
                ret = s.get(url, params=params)
                if ignore_errors and ret.status_code in ignore_errors:
                    break
                ret.raise_for_status()
                json = ret.json()

                # process results
                # FIXME: This assumes 'return type' matches a specific query,
                #        it will break if we expand the queries and they dont have different types
                if 'results' not in json:
                    # /hosts/:id dos not have a 'results' key
                    results = json
                    break
                elif isinstance(json['results'], MutableMapping):
                    # /facts are returned as dict in 'results'
                    results = json['results']
                    break
                else:
                    # /hosts 's 'results' is a list of all hosts, returned is paginated
                    results = results + json['results']

                    # check for end of paging
                    if len(results) >= json['subtotal']:
                        break
                    if len(json['results']) == 0:
                        self.display.warning("Did not make any progress during loop. expected %d got %d" % (json['subtotal'], len(results)))
                        break

                    # get next page
                    params['page'] += 1

            self._cache[self.cache_key][url] = results

        return self._cache[self.cache_key][url]

    def _get_hosts(self):
        return self._get_json("%s/api/v2/hosts" % self.foreman_url)

    def _get_all_params_by_id(self, hid):
        url = "%s/api/v2/hosts/%s" % (self.foreman_url, hid)
        ret = self._get_json(url, [404])
        if not ret or not isinstance(ret, MutableMapping) or not ret.get('all_parameters', False):
            return {}
        return ret.get('all_parameters')

    def _get_facts_by_id(self, hid):
        url = "%s/api/v2/hosts/%s/facts" % (self.foreman_url, hid)
        return self._get_json(url)

    def _get_facts(self, host):
        """Fetch all host facts of the host"""

        ret = self._get_facts_by_id(host['id'])
        if len(ret.values()) == 0:
            facts = {}
        elif len(ret.values()) == 1:
            facts = list(ret.values())[0]
        else:
            raise ValueError("More than one set of facts returned for '%s'" % host)
        return facts

    def _populate(self):

        for host in self._get_hosts():

            if host.get('name'):
                self.inventory.add_host(host['name'])

                # create directly mapped groups
                group_name = host.get('hostgroup_title', host.get('hostgroup_name'))
                if group_name:
                    group_name = to_safe_group_name('%s%s' % (self.get_option('group_prefix'), group_name.lower().replace(" ", "")))
                    group_name = self.inventory.add_group(group_name)
                    self.inventory.add_child(group_name, host['name'])

                # set host vars from host info
                try:
                    for k, v in host.items():
                        if k not in ('name', 'hostgroup_title', 'hostgroup_name'):
                            try:
                                self.inventory.set_variable(host['name'], self.get_option('vars_prefix') + k, v)
                            except ValueError as e:
                                self.display.warning("Could not set host info hostvar for %s, skipping %s: %s" % (host, k, to_text(e)))
                except ValueError as e:
                    self.display.warning("Could not get host info for %s, skipping: %s" % (host['name'], to_text(e)))

                # set host vars from params
                if self.get_option('want_params'):
                    for p in self._get_all_params_by_id(host['id']):
                        try:
                            self.inventory.set_variable(host['name'], p['name'], p['value'])
                        except ValueError as e:
                            self.display.warning("Could not set hostvar %s to '%s' for the '%s' host, skipping:  %s" %
                                                 (p['name'], to_native(p['value']), host, to_native(e)))

                # set host vars from facts
                if self.get_option('want_facts'):
                    self.inventory.set_variable(host['name'], 'ansible_facts', self._get_facts(host))

    def parse(self, inventory, loader, path, cache=True):

        super(InventoryModule, self).parse(inventory, loader, path)

        # read config from file, this sets 'options'
        self._read_config_data(path)

        # get connection host
        self.foreman_url = self.get_option('url')
        self.cache_key = self.get_cache_key(path)
        self.use_cache = cache and self.get_option('cache')

        # actually populate inventory
        self._populate()



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import datetime
import json

from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.module_utils.common._collections_compat import Set
from ansible.module_utils.six import (
    binary_type,
    iteritems,
    text_type,
)


def _json_encode_fallback(obj):
    if isinstance(obj, Set):
        return list(obj)
    elif isinstance(obj, datetime.datetime):
        return obj.isoformat()
    raise TypeError("Cannot json serialize %s" % to_native(obj))


def jsonify(data, **kwargs):
    for encoding in ("utf-8", "latin-1"):
        try:
            return json.dumps(data, encoding=encoding, default=_json_encode_fallback, **kwargs)
        # Old systems using old simplejson module does not support encoding keyword.
        except TypeError:
            try:
                new_data = container_to_text(data, encoding=encoding)
            except UnicodeDecodeError:
                continue
            return json.dumps(new_data, default=_json_encode_fallback, **kwargs)
        except UnicodeDecodeError:
            continue
    raise UnicodeError('Invalid unicode encoding encountered')


def container_to_bytes(d, encoding='utf-8', errors='surrogate_or_strict'):
    ''' Recursively convert dict keys and values to byte str

        Specialized for json return because this only handles, lists, tuples,
        and dict container types (the containers that the json module returns)
    '''

    if isinstance(d, text_type):
        return to_bytes(d, encoding=encoding, errors=errors)
    elif isinstance(d, dict):
        return dict(container_to_bytes(o, encoding, errors) for o in iteritems(d))
    elif isinstance(d, list):
        return [container_to_bytes(o, encoding, errors) for o in d]
    elif isinstance(d, tuple):
        return tuple(container_to_bytes(o, encoding, errors) for o in d)
    else:
        return d


def container_to_text(d, encoding='utf-8', errors='surrogate_or_strict'):
    """Recursively convert dict keys and values to byte str

    Specialized for json return because this only handles, lists, tuples,
    and dict container types (the containers that the json module returns)
    """

    if isinstance(d, binary_type):
        # Warning, can traceback
        return to_text(d, encoding=encoding, errors=errors)
    elif isinstance(d, dict):
        return dict(container_to_text(o, encoding, errors) for o in iteritems(d))
    elif isinstance(d, list):
        return [container_to_text(o, encoding, errors) for o in d]
    elif isinstance(d, tuple):
        return tuple(container_to_text(o, encoding, errors) for o in d)
    else:
        return d



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)#!/usr/bin/python
"""
The nxos lag_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""
from __future__ import absolute_import, division, print_function
__metaclass__ = type

import re
from copy import deepcopy

from ansible.module_utils.network.common import utils
from ansible.module_utils.network.nxos.argspec.lag_interfaces.lag_interfaces import Lag_interfacesArgs
from ansible.module_utils.network.nxos.utils.utils import get_interface_type, normalize_interface


class Lag_interfacesFacts(object):
    """ The nxos lag_interfaces fact class
    """

    def __init__(self, module, subspec='config', options='options'):
        self._module = module
        self.argument_spec = Lag_interfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def populate_facts(self, connection, ansible_facts, data=None):
        """ Populate the facts for lag_interfaces
        :param connection: the device connection
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        objs = []
        if not data:
            data = connection.get('show running-config | include channel-group')
        config = re.split('(\n  |)channel-group ', data)
        config = list(dict.fromkeys(config))
        for conf in config:
            if conf:
                obj = self.render_config(self.generated_spec, conf, connection)
                if obj and len(obj.keys()) > 1:
                    objs.append(obj)

        ansible_facts['ansible_network_resources'].pop('lag_interfaces', None)
        facts = {}
        if objs:
            facts['lag_interfaces'] = []
            params = utils.validate_config(self.argument_spec, {'config': objs})
            for cfg in params['config']:
                facts['lag_interfaces'].append(utils.remove_empties(cfg))

        ansible_facts['ansible_network_resources'].update(facts)
        return ansible_facts

    def get_members(self, id, connection):
        """
        Returns members associated with a channel-group

        :param name: The channel group
        :rtype: list
        :returns: Members
        """
        members = []
        data = connection.get('show port-channel summary')
        match = re.search(r'{0} (.+)(|\n)'.format(id), data)
        if match:
            interfaces = re.search(r'Eth\d(.+)$', match.group())
            if interfaces:
                for i in interfaces.group().split():
                    if get_interface_type(i[:-3]) != 'unknown':
                        members.append(normalize_interface(i[:-3]))

        return members

    def render_config(self, spec, conf, connection):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        match = re.search(r'(\d+)( |)(force )?(mode \S+)?', conf, re.M)
        if match:
            matches = match.groups()
            config['name'] = 'port-channel' + str(matches[0])
            config['members'] = []
            members = self.get_members(config['name'].strip('port-channel'), connection)
            if members:
                for m in members:
                    m_dict = {}
                    if matches[2]:
                        m_dict['force'] = matches[2]
                    if matches[3]:
                        m_dict['mode'] = matches[3][5:]
                    m_dict['member'] = m
                    config['members'].append(m_dict)
        else:
            config = {}

        lag_intf_cfg = utils.remove_empties(config)
        # if lag interfaces config is not present return empty dict
        if len(lag_intf_cfg) == 1:
            return {}
        else:
            return lag_intf_cfg



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Abhijeet Kasurde <akasurde@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['deprecated'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: vmware_host_ssl_facts
deprecated:
  removed_in: '2.13'
  why: Deprecated in favour of C(_info) module.
  alternative: Use M(vmware_host_ssl_info) instead.
short_description: Gather facts of ESXi host system about SSL
description:
- This module can be used to gather facts of the SSL thumbprint information for a host.
version_added: 2.7
author:
- Abhijeet Kasurde (@Akasurde)
notes:
- Tested on vSphere 6.5
requirements:
- python >= 2.6
- PyVmomi
options:
  cluster_name:
    description:
    - Name of the cluster.
    - SSL thumbprint information about all ESXi host system in the given cluster will be reported.
    - If C(esxi_hostname) is not given, this parameter is required.
    type: str
  esxi_hostname:
    description:
    - ESXi hostname.
    - SSL thumbprint information of this ESXi host system will be reported.
    - If C(cluster_name) is not given, this parameter is required.
    type: str
extends_documentation_fragment: vmware.documentation
'''

EXAMPLES = r'''
- name: Gather SSL thumbprint information about all ESXi Hosts in given Cluster
  vmware_host_ssl_facts:
    hostname: '{{ vcenter_hostname }}'
    username: '{{ vcenter_username }}'
    password: '{{ vcenter_password }}'
    cluster_name: '{{ cluster_name }}'
  delegate_to: localhost
  register: all_host_ssl_facts

- name: Get SSL Thumbprint info about "{{ esxi_hostname }}"
  vmware_host_ssl_facts:
    hostname: "{{ vcenter_server }}"
    username: "{{ vcenter_user }}"
    password: "{{ vcenter_pass }}"
    esxi_hostname: '{{ esxi_hostname }}'
  register: ssl_facts
- set_fact:
    ssl_thumbprint: "{{ ssl_facts['host_ssl_facts'][esxi_hostname]['ssl_thumbprints'][0] }}"
- debug:
    msg: "{{ ssl_thumbprint }}"
- name: Add ESXi Host to vCenter
  vmware_host:
    hostname: '{{ vcenter_hostname }}'
    username: '{{ vcenter_username }}'
    password: '{{ vcenter_password }}'
    datacenter_name: '{{ datacenter_name }}'
    cluster_name: '{{ cluster_name }}'
    esxi_hostname: '{{ esxi_hostname }}'
    esxi_username: '{{ esxi_username }}'
    esxi_password: '{{ esxi_password }}'
    esxi_ssl_thumbprint: '{{ ssl_thumbprint }}'
    state: present
'''

RETURN = r'''
host_ssl_facts:
    description:
    - dict with hostname as key and dict with SSL thumbprint related facts
    returned: facts
    type: dict
    sample:
        {
            "10.76.33.215": {
                "owner_tag": "",
                "principal": "vpxuser",
                "ssl_thumbprints": [
                    "E3:E8:A9:20:8D:32:AE:59:C6:8D:A5:91:B0:20:EF:00:A2:7C:27:EE",
                    "F1:AC:DA:6E:D8:1E:37:36:4A:5C:07:E5:04:0B:87:C8:75:FB:42:01"
                ]
            }
        }
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.vmware import vmware_argument_spec, PyVmomi


class VMwareHostSslManager(PyVmomi):
    def __init__(self, module):
        super(VMwareHostSslManager, self).__init__(module)
        cluster_name = self.params.get('cluster_name', None)
        esxi_host_name = self.params.get('esxi_hostname', None)
        self.hosts = self.get_all_host_objs(cluster_name=cluster_name, esxi_host_name=esxi_host_name)
        self.hosts_facts = {}

    def gather_ssl_facts(self):
        for host in self.hosts:
            self.hosts_facts[host.name] = dict(principal='',
                                               owner_tag='',
                                               ssl_thumbprints=[])

            host_ssl_info_mgr = host.config.sslThumbprintInfo
            if host_ssl_info_mgr:
                self.hosts_facts[host.name]['principal'] = host_ssl_info_mgr.principal
                self.hosts_facts[host.name]['owner_tag'] = host_ssl_info_mgr.ownerTag
                self.hosts_facts[host.name]['ssl_thumbprints'] = [i for i in host_ssl_info_mgr.sslThumbprints]

        self.module.exit_json(changed=False, host_ssl_facts=self.hosts_facts)


def main():
    argument_spec = vmware_argument_spec()
    argument_spec.update(
        cluster_name=dict(type='str'),
        esxi_hostname=dict(type='str'),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_one_of=[
            ['cluster_name', 'esxi_hostname'],
        ],
        supports_check_mode=True,
    )

    vmware_host_accept_config = VMwareHostSslManager(module)
    vmware_host_accept_config.gather_ssl_facts()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Joseph Callen <jcallen () csc.com>
# Copyright: (c) 2017, Ansible Project
# Copyright: (c) 2017, Abhijeet Kasurde <akasurde@redhat.com>
# Copyright: (c) 2018, Christian Kotte <christian.kotte@gmx.de>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'community'
}

DOCUMENTATION = '''
---
module: vmware_portgroup
short_description: Create a VMware portgroup
description:
    - Create a VMware Port Group on a VMware Standard Switch (vSS) for given ESXi host(s) or hosts of given cluster.
version_added: 2.0
author:
- Joseph Callen (@jcpowermac)
- Russell Teague (@mtnbikenc)
- Abhijeet Kasurde (@Akasurde)
- Christian Kotte (@ckotte)
notes:
    - Tested on vSphere 5.5 and 6.5
    - Complete configuration only tested on vSphere 6.5
    - 'C(inbound_policy) and C(rolling_order) are deprecated and will be removed in 2.11'
    - Those two options are only used during portgroup creation. Updating isn't supported with those options.
requirements:
    - "python >= 2.6"
    - PyVmomi
options:
    switch:
        description:
            - vSwitch to modify.
        required: True
        aliases: [ 'switch_name', 'vswitch' ]
        type: str
    portgroup:
        description:
            - Portgroup name to add.
        required: True
        aliases: [ 'portgroup_name' ]
        type: str
    vlan_id:
        description:
            - VLAN ID to assign to portgroup.
            - Set to 0 (no VLAN tagging) by default.
        required: False
        default: 0
        aliases: [ 'vlan' ]
        type: int
    security:
        description:
            - Network policy specifies layer 2 security settings for a
              portgroup such as promiscuous mode, where guest adapter listens
              to all the packets, MAC address changes and forged transmits.
            - Dict which configures the different security values for portgroup.
            - 'Valid attributes are:'
            - '- C(promiscuous_mode) (bool): indicates whether promiscuous mode is allowed. (default: None)'
            - '- C(forged_transmits) (bool): indicates whether forged transmits are allowed. (default: None)'
            - '- C(mac_changes) (bool): indicates whether mac changes are allowed. (default: None)'
        required: False
        version_added: "2.2"
        aliases: [ 'security_policy', 'network_policy' ]
        type: dict
    teaming:
        description:
            - Dictionary which configures the different teaming values for portgroup.
            - 'Valid attributes are:'
            - '- C(load_balancing) (string): Network adapter teaming policy. C(load_balance_policy) is also alias to this option. (default: loadbalance_srcid)'
            - '   - choices: [ loadbalance_ip, loadbalance_srcmac, loadbalance_srcid, failover_explicit ]'
            - '- C(network_failure_detection) (string): Network failure detection. (default: link_status_only)'
            - '   - choices: [ link_status_only, beacon_probing ]'
            - '- C(notify_switches) (bool): Indicate whether or not to notify the physical switch if a link fails. (default: None)'
            - '- C(failback) (bool): Indicate whether or not to use a failback when restoring links. (default: None)'
            - '- C(active_adapters) (list): List of active adapters used for load balancing.'
            - '- C(standby_adapters) (list): List of standby adapters used for failover.'
            - '- All vmnics are used as active adapters if C(active_adapters) and C(standby_adapters) are not defined.'
            - '- C(inbound_policy) (bool): Indicate whether or not the teaming policy is applied to inbound frames as well. Deprecated. (default: False)'
            - '- C(rolling_order) (bool): Indicate whether or not to use a rolling policy when restoring links. Deprecated. (default: False)'
        required: False
        version_added: '2.6'
        aliases: [ 'teaming_policy' ]
        type: dict
    traffic_shaping:
        description:
            - Dictionary which configures traffic shaping for the switch.
            - 'Valid attributes are:'
            - '- C(enabled) (bool): Status of Traffic Shaping Policy. (default: None)'
            - '- C(average_bandwidth) (int): Average bandwidth (kbit/s). (default: None)'
            - '- C(peak_bandwidth) (int): Peak bandwidth (kbit/s). (default: None)'
            - '- C(burst_size) (int): Burst size (KB). (default: None)'
        required: False
        version_added: '2.9'
        type: dict
    cluster_name:
        description:
            - Name of cluster name for host membership.
            - Portgroup will be created on all hosts of the given cluster.
            - This option is required if C(hosts) is not specified.
        version_added: "2.5"
        aliases: [ 'cluster' ]
        type: str
    hosts:
        description:
            - List of name of host or hosts on which portgroup needs to be added.
            - This option is required if C(cluster_name) is not specified.
        aliases: [ esxi_hostname ]
        version_added: "2.5"
        type: list
    state:
        description:
            - Determines if the portgroup should be present or not.
        choices:
            - 'present'
            - 'absent'
        version_added: '2.5'
        default: present
        type: str
extends_documentation_fragment: vmware.documentation
'''

EXAMPLES = r'''
- name: Add Management Network VM Portgroup
  vmware_portgroup:
    hostname: "{{ esxi_hostname }}"
    username: "{{ esxi_username }}"
    password: "{{ esxi_password }}"
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    vlan_id: "{{ vlan_id }}"
  delegate_to: localhost

- name: Add Portgroup with Promiscuous Mode Enabled
  vmware_portgroup:
    hostname: "{{ esxi_hostname }}"
    username: "{{ esxi_username }}"
    password: "{{ esxi_password }}"
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    security:
        promiscuous_mode: True
  delegate_to: localhost

- name: Add Management Network VM Portgroup to specific hosts
  vmware_portgroup:
    hostname: "{{ vcenter_hostname }}"
    username: "{{ vcenter_username }}"
    password: "{{ vcenter_password }}"
    hosts: [esxi_hostname_one]
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    vlan_id: "{{ vlan_id }}"
  delegate_to: localhost

- name: Add Management Network VM Portgroup to all hosts in a cluster
  vmware_portgroup:
    hostname: "{{ vcenter_hostname }}"
    username: "{{ vcenter_username }}"
    password: "{{ vcenter_password }}"
    cluster_name: "{{ cluster_name }}"
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    vlan_id: "{{ vlan_id }}"
  delegate_to: localhost

- name: Remove Management Network VM Portgroup to all hosts in a cluster
  vmware_portgroup:
    hostname: "{{ vcenter_hostname }}"
    username: "{{ vcenter_username }}"
    password: "{{ vcenter_password }}"
    cluster_name: "{{ cluster_name }}"
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    vlan_id: "{{ vlan_id }}"
    state: absent
  delegate_to: localhost

- name: Add Portgroup with all settings defined
  vmware_portgroup:
    hostname: "{{ vcenter_hostname }}"
    username: "{{ vcenter_username }}"
    password: "{{ vcenter_password }}"
    esxi_hostname: "{{ inventory_hostname }}"
    switch: "{{ vswitch_name }}"
    portgroup: "{{ portgroup_name }}"
    vlan_id: 10
    security:
        promiscuous_mode: False
        mac_changes: False
        forged_transmits: False
    traffic_shaping:
        enabled: True
        average_bandwidth: 100000
        peak_bandwidth: 100000
        burst_size: 102400
    teaming:
        load_balancing: failover_explicit
        network_failure_detection: link_status_only
        notify_switches: true
        failback: true
        active_adapters:
            - vmnic0
        standby_adapters:
            - vmnic1
  delegate_to: localhost
  register: teaming_result
'''

RETURN = r'''
result:
    description: metadata about the portgroup
    returned: always
    type: dict
    sample: {
        "esxi01.example.com": {
            "changed": true,
            "failback": "No override",
            "failover_active": "No override",
            "failover_standby": "No override",
            "failure_detection": "No override",
            "load_balancing": "No override",
            "msg": "Port Group added",
            "notify_switches": "No override",
            "portgroup": "vMotion",
            "sec_forged_transmits": false,
            "sec_mac_changes": false,
            "sec_promiscuous_mode": false,
            "traffic_shaping": "No override",
            "vlan_id": 33,
            "vswitch": "vSwitch1"
        }
    }
'''


try:
    from pyVmomi import vim, vmodl
except ImportError:
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.vmware import PyVmomi, vmware_argument_spec
from ansible.module_utils._text import to_native


class VMwareHostPortGroup(PyVmomi):
    """Manage portgroup"""

    def __init__(self, module):
        super(VMwareHostPortGroup, self).__init__(module)
        self.switch_object = None
        self.portgroup_object = None
        hosts = self.params['hosts']
        cluster = self.params['cluster_name']
        self.portgroup = self.params['portgroup']
        self.switch = self.params['switch']
        self.vlan_id = self.params['vlan_id']
        if self.params['security']:
            self.sec_promiscuous_mode = self.params['security'].get('promiscuous_mode')
            self.sec_forged_transmits = self.params['security'].get('forged_transmits')
            self.sec_mac_changes = self.params['security'].get('mac_changes')
        else:
            self.sec_promiscuous_mode = None
            self.sec_forged_transmits = None
            self.sec_mac_changes = None
        if self.params['traffic_shaping']:
            self.ts_enabled = self.params['traffic_shaping'].get('enabled')
            for value in ['average_bandwidth', 'peak_bandwidth', 'burst_size']:
                if not self.params['traffic_shaping'].get(value):
                    self.module.fail_json(msg="traffic_shaping.%s is a required parameter if traffic_shaping is enabled." % value)
            self.ts_average_bandwidth = self.params['traffic_shaping'].get('average_bandwidth')
            self.ts_peak_bandwidth = self.params['traffic_shaping'].get('peak_bandwidth')
            self.ts_burst_size = self.params['traffic_shaping'].get('burst_size')
        else:
            self.ts_enabled = None
            self.ts_average_bandwidth = None
            self.ts_peak_bandwidth = None
            self.ts_burst_size = None
        if self.params['teaming']:
            self.teaming_load_balancing = self.params['teaming'].get('load_balancing')
            self.teaming_failure_detection = self.params['teaming'].get('network_failure_detection')
            self.teaming_notify_switches = self.params['teaming'].get('notify_switches')
            self.teaming_failback = self.params['teaming'].get('failback')
            self.teaming_failover_order_active = self.params['teaming'].get('active_adapters')
            self.teaming_failover_order_standby = self.params['teaming'].get('standby_adapters')
            if self.teaming_failover_order_active is None:
                self.teaming_failover_order_active = []
            if self.teaming_failover_order_standby is None:
                self.teaming_failover_order_standby = []
            # NOTE: the following options are deprecated and should be removed in 2.11
            self.teaming_inbound_policy = self.module.params['teaming']['inbound_policy']
            self.teaming_rolling_order = self.module.params['teaming']['rolling_order']
        else:
            self.teaming_load_balancing = None
            self.teaming_failure_detection = None
            self.teaming_notify_switches = None
            self.teaming_failback = None
            self.teaming_failover_order_active = None
            self.teaming_failover_order_standby = None
            # NOTE: the following options are deprecated and should be removed in 2.11
            self.teaming_inbound_policy = None
            self.teaming_rolling_order = None
        self.state = self.params['state']

        self.hosts = self.get_all_host_objs(cluster_name=cluster, esxi_host_name=hosts)
        if not self.hosts:
            self.module.fail_json(msg="Failed to find host system with given configuration.")

    def process_state(self):
        """Manage internal state of the portgroup"""
        results = dict(changed=False, result=dict())
        host_change_list = []
        for host in self.hosts:
            changed = False
            results['result'][host.name] = dict()
            switch_state = self.check_if_vswitch_exists(host_system=host)
            if switch_state == 'absent':
                self.module.fail_json(msg="The vSwitch '%s' doesn't exist on host '%s'" % (self.switch, host.name))
            portgroup_state = self.check_if_portgroup_exists(host_system=host)
            if self.state == 'present' and portgroup_state == 'present':
                changed, host_results = self.update_host_port_group(
                    host_system=host,
                    portgroup_object=self.portgroup_object
                )
            elif self.state == 'present' and portgroup_state == 'absent':
                changed, host_results = self.create_host_port_group(host_system=host)
            elif self.state == 'absent' and portgroup_state == 'present':
                changed, host_results = self.remove_host_port_group(host_system=host)
            else:
                host_results = dict()
                host_results['changed'] = False
                host_results['msg'] = "Port Group already deleted"
                host_results['portgroup'] = self.portgroup
            results['result'][host.name] = host_results

            host_change_list.append(changed)

        if any(host_change_list):
            results['changed'] = True
        self.module.exit_json(**results)

    def check_if_portgroup_exists(self, host_system):
        """
        Check if portgroup exists
        Returns: 'present' if portgroup exists or 'absent' if not
        """
        self.portgroup_object = self.find_portgroup_by_name(
            host_system=host_system,
            portgroup_name=self.portgroup,
            vswitch_name=self.switch
        )
        if self.portgroup_object is None:
            return 'absent'
        return 'present'

    def find_portgroup_by_name(self, host_system, portgroup_name, vswitch_name):
        """
        Find and return port group managed object
        Args:
            host_system: Name of Host System
            portgroup_name: Name of the Port Group
            vswitch_name: Name of the vSwitch

        Returns: Port Group managed object if found, else None
        """
        portgroups = self.get_all_port_groups_by_host(host_system=host_system)
        for portgroup in portgroups:
            if portgroup.spec.name == portgroup_name and portgroup.spec.vswitchName != vswitch_name:
                # portgroup names are unique; there can be only one portgroup with the same name per host
                self.module.fail_json(msg="The portgroup already exists on vSwitch '%s'" % portgroup.spec.vswitchName)
            if portgroup.spec.name == portgroup_name and portgroup.spec.vswitchName == vswitch_name:
                return portgroup
        return None

    def check_if_vswitch_exists(self, host_system):
        """
        Check if vSwitch exists
        Returns: 'present' if vSwitch exists or 'absent' if not
        """
        self.switch_object = self.find_vswitch_by_name(
            host_system=host_system,
            vswitch_name=self.switch
        )
        if self.switch_object is None:
            return 'absent'
        return 'present'

    @staticmethod
    def find_vswitch_by_name(host_system, vswitch_name):
        """
        Find and return vSwitch managed object
        Args:
            host: Host system managed object
            vswitch_name: Name of vSwitch to find

        Returns: vSwitch managed object if found, else None
        """
        for vss in host_system.configManager.networkSystem.networkInfo.vswitch:
            if vss.name == vswitch_name:
                return vss
        return None

    def remove_host_port_group(self, host_system):
        """
        Remove a Port Group from a given host
        Args:
            host_system: Name of Host System
        """
        host_results = dict(changed=False, msg="")

        if self.module.check_mode:
            host_results['msg'] = "Port Group would be removed"
        else:
            try:
                host_system.configManager.networkSystem.RemovePortGroup(pgName=self.portgroup)
                host_results['msg'] = "Port Group removed"
            except vim.fault.NotFound as not_found:
                self.module.fail_json(
                    msg="Failed to remove Portgroup as it was not found: %s" % to_native(not_found.msg)
                )
            except vim.fault.ResourceInUse as resource_in_use:
                self.module.fail_json(
                    msg="Failed to remove Portgroup as it is in use: %s" % to_native(resource_in_use.msg)
                )
            except vim.fault.HostConfigFault as host_config_fault:
                self.module.fail_json(
                    msg="Failed to remove Portgroup due to configuration failures: %s" % to_native(host_config_fault.msg)
                )
        host_results['changed'] = True
        host_results['portgroup'] = self.portgroup
        host_results['vswitch'] = self.switch

        return True, host_results

    def create_host_port_group(self, host_system):
        """Create Port Group on a given host
        Args:
            host_system: Name of Host System
        """
        host_results = dict(changed=False, msg="")

        if self.module.check_mode:
            host_results['msg'] = "Port Group would be added"
        else:
            port_group = vim.host.PortGroup.Config()
            port_group.spec = vim.host.PortGroup.Specification()
            port_group.spec.vswitchName = self.switch
            port_group.spec.name = self.portgroup
            port_group.spec.vlanId = self.vlan_id
            port_group.spec.policy = self.create_network_policy()

            try:
                host_system.configManager.networkSystem.AddPortGroup(portgrp=port_group.spec)
                host_results['changed'] = True
                host_results['msg'] = "Port Group added"
            except vim.fault.AlreadyExists as already_exists:
                self.module.fail_json(
                    msg="Failed to add Portgroup as it already exists: %s" % to_native(already_exists.msg)
                )
            except vim.fault.NotFound as not_found:
                self.module.fail_json(
                    msg="Failed to add Portgroup as vSwitch was not found: %s" % to_native(not_found.msg)
                )
            except vim.fault.HostConfigFault as host_config_fault:
                self.module.fail_json(
                    msg="Failed to add Portgroup due to host system configuration failure : %s" %
                    to_native(host_config_fault.msg)
                )
            except vmodl.fault.InvalidArgument as invalid_argument:
                self.module.fail_json(
                    msg="Failed to add Portgroup as VLAN id was not correct as per specifications: %s" %
                    to_native(invalid_argument.msg)
                )
        host_results['changed'] = True
        host_results['portgroup'] = self.portgroup
        host_results['vswitch'] = self.switch
        host_results['vlan_id'] = self.vlan_id
        if self.sec_promiscuous_mode is None:
            host_results['sec_promiscuous_mode'] = "No override"
        else:
            host_results['sec_promiscuous_mode'] = self.sec_promiscuous_mode
        if self.sec_mac_changes is None:
            host_results['sec_mac_changes'] = "No override"
        else:
            host_results['sec_mac_changes'] = self.sec_mac_changes
        if self.sec_forged_transmits is None:
            host_results['sec_forged_transmits'] = "No override"
        else:
            host_results['sec_forged_transmits'] = self.sec_forged_transmits
        host_results['traffic_shaping'] = "No override" if self.ts_enabled is None else self.ts_enabled
        host_results['load_balancing'] = "No override" if self.teaming_load_balancing is None \
            else self.teaming_load_balancing
        host_results['notify_switches'] = "No override" if self.teaming_notify_switches is None \
            else self.teaming_notify_switches
        host_results['failback'] = "No override" if self.teaming_failback is None else self.teaming_failback
        host_results['failover_active'] = "No override" if self.teaming_failover_order_active is None \
            else self.teaming_failover_order_active
        host_results['failover_standby'] = "No override" if self.teaming_failover_order_standby is None \
            else self.teaming_failover_order_standby
        host_results['failure_detection'] = "No override" if self.teaming_failure_detection is None \
            else self.teaming_failure_detection

        return True, host_results

    def update_host_port_group(self, host_system, portgroup_object):
        """Update a Port Group on a given host
        Args:
            host_system: Name of Host System
        """
        changed = changed_security = False
        changed_list = []
        host_results = dict(changed=False, msg="")
        spec = portgroup_object.spec

        # Check VLAN ID
        host_results['vlan_id'] = self.vlan_id
        if spec.vlanId != self.vlan_id:
            changed = True
            changed_list.append("VLAN ID")
            host_results['vlan_id_previous'] = spec.vlanId
            spec.vlanId = self.vlan_id

        # Check security settings
        if self.sec_promiscuous_mode is None:
            host_results['sec_promiscuous_mode'] = "No override"
        else:
            host_results['sec_promiscuous_mode'] = self.sec_promiscuous_mode
        if self.sec_mac_changes is None:
            host_results['sec_mac_changes'] = "No override"
        else:
            host_results['sec_mac_changes'] = self.sec_mac_changes
        if self.sec_forged_transmits is None:
            host_results['sec_forged_transmits'] = "No override"
        else:
            host_results['sec_forged_transmits'] = self.sec_forged_transmits
        if spec.policy.security:
            promiscuous_mode_previous = spec.policy.security.allowPromiscuous
            mac_changes_previous = spec.policy.security.macChanges
            forged_transmits_previous = spec.policy.security.forgedTransmits
            if promiscuous_mode_previous is not self.sec_promiscuous_mode:
                spec.policy.security.allowPromiscuous = self.sec_promiscuous_mode
                changed = changed_security = True
                changed_list.append("Promiscuous mode")
            if mac_changes_previous is not self.sec_mac_changes:
                spec.policy.security.macChanges = self.sec_mac_changes
                changed = changed_security = True
                changed_list.append("MAC address changes")
            if forged_transmits_previous is not self.sec_forged_transmits:
                spec.policy.security.forgedTransmits = self.sec_forged_transmits
                changed = changed_security = True
                changed_list.append("Forged transmits")
            if changed_security:
                if self.sec_promiscuous_mode is None:
                    host_results['sec_promiscuous_mode_previous'] = "No override"
                else:
                    host_results['sec_promiscuous_mode_previous'] = promiscuous_mode_previous
                if self.sec_mac_changes is None:
                    host_results['sec_mac_changes_previous'] = "No override"
                else:
                    host_results['sec_mac_changes'] = mac_changes_previous
                if self.sec_forged_transmits is None:
                    host_results['sec_forged_transmits_previous'] = "No override"
                else:
                    host_results['sec_forged_transmits_previous'] = forged_transmits_previous
        else:
            spec.policy.security = self.create_security_policy()
            changed = True
            changed_list.append("Security")
            host_results['sec_promiscuous_mode_previous'] = "No override"
            host_results['sec_mac_changes_previous'] = "No override"
            host_results['sec_forged_transmits_previous'] = "No override"

        # Check traffic shaping
        if self.ts_enabled is None:
            host_results['traffic_shaping'] = "No override"
        else:
            host_results['traffic_shaping'] = self.ts_enabled
            if self.ts_enabled:
                ts_average_bandwidth = self.ts_average_bandwidth * 1000
                ts_peak_bandwidth = self.ts_peak_bandwidth * 1000
                ts_burst_size = self.ts_burst_size * 1024
                host_results['traffic_shaping_avg_bandw'] = ts_average_bandwidth
                host_results['traffic_shaping_peak_bandw'] = ts_peak_bandwidth
                host_results['traffic_shaping_burst'] = ts_burst_size
        if spec.policy.shapingPolicy and spec.policy.shapingPolicy.enabled is not None:
            if spec.policy.shapingPolicy.enabled:
                if self.ts_enabled:
                    if spec.policy.shapingPolicy.averageBandwidth != ts_average_bandwidth:
                        changed = True
                        changed_list.append("Average bandwidth")
                        host_results['traffic_shaping_avg_bandw_previous'] = spec.policy.shapingPolicy.averageBandwidth
                        spec.policy.shapingPolicy.averageBandwidth = ts_average_bandwidth
                    if spec.policy.shapingPolicy.peakBandwidth != ts_peak_bandwidth:
                        changed = True
                        changed_list.append("Peak bandwidth")
                        host_results['traffic_shaping_peak_bandw_previous'] = spec.policy.shapingPolicy.peakBandwidth
                        spec.policy.shapingPolicy.peakBandwidth = ts_peak_bandwidth
                    if spec.policy.shapingPolicy.burstSize != ts_burst_size:
                        changed = True
                        changed_list.append("Burst size")
                        host_results['traffic_shaping_burst_previous'] = spec.policy.shapingPolicy.burstSize
                        spec.policy.shapingPolicy.burstSize = ts_burst_size
                elif self.ts_enabled is False:
                    changed = True
                    changed_list.append("Traffic shaping")
                    host_results['traffic_shaping_previous'] = True
                    spec.policy.shapingPolicy.enabled = False
                elif self.ts_enabled is None:
                    spec.policy.shapingPolicy = None
                    changed = True
                    changed_list.append("Traffic shaping")
                    host_results['traffic_shaping_previous'] = True
            else:
                if self.ts_enabled:
                    spec.policy.shapingPolicy = self.create_shaping_policy()
                    changed = True
                    changed_list.append("Traffic shaping")
                    host_results['traffic_shaping_previous'] = False
                elif self.ts_enabled is False:
                    changed = True
                    changed_list.append("Traffic shaping")
                    host_results['traffic_shaping_previous'] = True
                    spec.policy.shapingPolicy.enabled = False
                elif self.ts_enabled is None:
                    spec.policy.shapingPolicy = None
                    changed = True
                    changed_list.append("Traffic shaping")
                    host_results['traffic_shaping_previous'] = True
        else:
            if self.ts_enabled:
                spec.policy.shapingPolicy = self.create_shaping_policy()
                changed = True
                changed_list.append("Traffic shaping")
                host_results['traffic_shaping_previous'] = "No override"
            elif self.ts_enabled is False:
                changed = True
                changed_list.append("Traffic shaping")
                host_results['traffic_shaping_previous'] = "No override"
                spec.policy.shapingPolicy.enabled = False

        # Check teaming
        if spec.policy.nicTeaming:
            # Check teaming policy
            if self.teaming_load_balancing is None:
                host_results['load_balancing'] = "No override"
            else:
                host_results['load_balancing'] = self.teaming_load_balancing
            if spec.policy.nicTeaming.policy:
                if spec.policy.nicTeaming.policy != self.teaming_load_balancing:
                    changed = True
                    changed_list.append("Load balancing")
                    host_results['load_balancing_previous'] = spec.policy.nicTeaming.policy
                    spec.policy.nicTeaming.policy = self.teaming_load_balancing
            else:
                if self.teaming_load_balancing:
                    changed = True
                    changed_list.append("Load balancing")
                    host_results['load_balancing_previous'] = "No override"
                    spec.policy.nicTeaming.policy = self.teaming_load_balancing
            # Check teaming notify switches
            if spec.policy.nicTeaming.notifySwitches is None:
                host_results['notify_switches'] = "No override"
            else:
                host_results['notify_switches'] = self.teaming_notify_switches
            if spec.policy.nicTeaming.notifySwitches is not None:
                if self.teaming_notify_switches is not None:
                    if spec.policy.nicTeaming.notifySwitches is not self.teaming_notify_switches:
                        changed = True
                        changed_list.append("Notify switches")
                        host_results['notify_switches_previous'] = spec.policy.nicTeaming.notifySwitches
                        spec.policy.nicTeaming.notifySwitches = self.teaming_notify_switches
                else:
                    changed = True
                    changed_list.append("Notify switches")
                    host_results['notify_switches_previous'] = spec.policy.nicTeaming.notifySwitches
                    spec.policy.nicTeaming.notifySwitches = None
            else:
                if self.teaming_notify_switches is not None:
                    changed = True
                    changed_list.append("Notify switches")
                    host_results['notify_switches_previous'] = "No override"
                    spec.policy.nicTeaming.notifySwitches = self.teaming_notify_switches
            # Check failback
            if spec.policy.nicTeaming.rollingOrder is None:
                host_results['failback'] = "No override"
            else:
                host_results['failback'] = self.teaming_failback
            if spec.policy.nicTeaming.rollingOrder is not None:
                if self.teaming_failback is not None:
                    # this option is called 'failback' in the vSphere Client
                    # rollingOrder also uses the opposite value displayed in the client
                    if spec.policy.nicTeaming.rollingOrder is self.teaming_failback:
                        changed = True
                        changed_list.append("Failback")
                        host_results['failback_previous'] = not spec.policy.nicTeaming.rollingOrder
                        spec.policy.nicTeaming.rollingOrder = not self.teaming_failback
                else:
                    changed = True
                    changed_list.append("Failback")
                    host_results['failback_previous'] = spec.policy.nicTeaming.rollingOrder
                    spec.policy.nicTeaming.rollingOrder = None
            else:
                if self.teaming_failback is not None:
                    changed = True
                    changed_list.append("Failback")
                    host_results['failback_previous'] = "No override"
                    spec.policy.nicTeaming.rollingOrder = not self.teaming_failback
            # Check teaming failover order
            if self.teaming_failover_order_active is None and self.teaming_failover_order_standby is None:
                host_results['failover_active'] = "No override"
                host_results['failover_standby'] = "No override"
            else:
                host_results['failover_active'] = self.teaming_failover_order_active
                host_results['failover_standby'] = self.teaming_failover_order_standby
            if spec.policy.nicTeaming.nicOrder:
                if self.teaming_failover_order_active or self.teaming_failover_order_standby:
                    if spec.policy.nicTeaming.nicOrder.activeNic != self.teaming_failover_order_active:
                        changed = True
                        changed_list.append("Failover order active")
                        host_results['failover_active_previous'] = spec.policy.nicTeaming.nicOrder.activeNic
                        spec.policy.nicTeaming.nicOrder.activeNic = self.teaming_failover_order_active
                    if spec.policy.nicTeaming.nicOrder.standbyNic != self.teaming_failover_order_standby:
                        changed = True
                        changed_list.append("Failover order standby")
                        host_results['failover_standby_previous'] = spec.policy.nicTeaming.nicOrder.standbyNic
                        spec.policy.nicTeaming.nicOrder.standbyNic = self.teaming_failover_order_standby
                else:
                    spec.policy.nicTeaming.nicOrder = None
                    changed = True
                    changed_list.append("Failover order")
                    if hasattr(spec.policy.nicTeaming.nicOrder, 'activeNic'):
                        host_results['failover_active_previous'] = spec.policy.nicTeaming.nicOrder.activeNic
                    else:
                        host_results['failover_active_previous'] = []
                    if hasattr(spec.policy.nicTeaming.nicOrder, 'standbyNic'):
                        host_results['failover_standby_previous'] = spec.policy.nicTeaming.nicOrder.standbyNic
                    else:
                        host_results['failover_standby_previous'] = []
            else:
                if self.teaming_failover_order_active or self.teaming_failover_order_standby:
                    changed = True
                    changed_list.append("Failover order")
                    host_results['failover_active_previous'] = "No override"
                    host_results['failover_standby_previous'] = "No override"
                    spec.policy.nicTeaming.nicOrder = self.create_nic_order_policy()
            # Check teaming failure detection
            if self.teaming_failure_detection is None:
                host_results['failure_detection'] = "No override"
            else:
                host_results['failure_detection'] = self.teaming_failure_detection
            if spec.policy.nicTeaming.failureCriteria and spec.policy.nicTeaming.failureCriteria.checkBeacon is not None:
                if self.teaming_failure_detection == "link_status_only":
                    if spec.policy.nicTeaming.failureCriteria.checkBeacon is True:
                        changed = True
                        changed_list.append("Network failure detection")
                        host_results['failure_detection_previous'] = "beacon_probing"
                        spec.policy.nicTeaming.failureCriteria.checkBeacon = False
                elif self.teaming_failure_detection == "beacon_probing":
                    if spec.policy.nicTeaming.failureCriteria.checkBeacon is False:
                        changed = True
                        changed_list.append("Network failure detection")
                        host_results['failure_detection_previous'] = "link_status_only"
                        spec.policy.nicTeaming.failureCriteria.checkBeacon = True
                elif spec.policy.nicTeaming.failureCriteria.checkBeacon is not None:
                    changed = True
                    changed_list.append("Network failure detection")
                    host_results['failure_detection_previous'] = spec.policy.nicTeaming.failureCriteria.checkBeacon
                    spec.policy.nicTeaming.failureCriteria = None
            else:
                if self.teaming_failure_detection:
                    spec.policy.nicTeaming.failureCriteria = self.create_nic_failure_policy()
                    changed = True
                    changed_list.append("Network failure detection")
                    host_results['failure_detection_previous'] = "No override"
        else:
            spec.policy.nicTeaming = self.create_teaming_policy()
            if spec.policy.nicTeaming:
                changed = True
                changed_list.append("Teaming and failover")
                host_results['load_balancing_previous'] = "No override"
                host_results['notify_switches_previous'] = "No override"
                host_results['failback_previous'] = "No override"
                host_results['failover_active_previous'] = "No override"
                host_results['failover_standby_previous'] = "No override"
                host_results['failure_detection_previous'] = "No override"

        if changed:
            if self.module.check_mode:
                changed_suffix = ' would be changed'
            else:
                changed_suffix = ' changed'
            if len(changed_list) > 2:
                message = ', '.join(changed_list[:-1]) + ', and ' + str(changed_list[-1])
            elif len(changed_list) == 2:
                message = ' and '.join(changed_list)
            elif len(changed_list) == 1:
                message = changed_list[0]
            message += changed_suffix
            if not self.module.check_mode:
                try:
                    host_system.configManager.networkSystem.UpdatePortGroup(
                        pgName=self.portgroup,
                        portgrp=spec
                    )
                except vim.fault.AlreadyExists as already_exists:
                    self.module.fail_json(
                        msg="Failed to update Portgroup as it would conflict with an existing port group: %s" %
                        to_native(already_exists.msg)
                    )
                except vim.fault.NotFound as not_found:
                    self.module.fail_json(
                        msg="Failed to update Portgroup as vSwitch was not found: %s" %
                        to_native(not_found.msg)
                    )
                except vim.fault.HostConfigFault as host_config_fault:
                    self.module.fail_json(
                        msg="Failed to update Portgroup due to host system configuration failure : %s" %
                        to_native(host_config_fault.msg)
                    )
                except vmodl.fault.InvalidArgument as invalid_argument:
                    self.module.fail_json(
                        msg="Failed to update Port Group '%s', this can be due to either of following :"
                        " 1. VLAN id was not correct as per specifications, 2. Network policy is invalid : %s" %
                        (self.portgroup, to_native(invalid_argument.msg))
                    )
        else:
            message = "Port Group already configured properly"
        host_results['changed'] = changed
        host_results['msg'] = message
        host_results['portgroup'] = self.portgroup
        host_results['vswitch'] = self.switch

        return changed, host_results

    def create_network_policy(self):
        """
        Create a Network Policy
        Returns: Network Policy object
        """
        security_policy = None
        shaping_policy = None
        teaming_policy = None

        # Only configure security policy if an option is defined
        if not all(option is None for option in [self.sec_promiscuous_mode,
                                                 self.sec_mac_changes,
                                                 self.sec_forged_transmits]):
            security_policy = self.create_security_policy()
        if self.ts_enabled:
            shaping_policy = self.create_shaping_policy()
        teaming_policy = self.create_teaming_policy()

        network_policy = vim.host.NetworkPolicy(
            security=security_policy,
            nicTeaming=teaming_policy,
            shapingPolicy=shaping_policy
        )

        return network_policy

    def create_security_policy(self):
        """
        Create a Security Policy
        Returns: Security Policy object
        """
        security_policy = vim.host.NetworkPolicy.SecurityPolicy()
        security_policy.allowPromiscuous = self.sec_promiscuous_mode
        security_policy.macChanges = self.sec_mac_changes
        security_policy.forgedTransmits = self.sec_forged_transmits
        return security_policy

    def create_shaping_policy(self):
        """
        Create a Traffic Shaping Policy
        Returns: Traffic Shaping Policy object
        """
        shaping_policy = vim.host.NetworkPolicy.TrafficShapingPolicy()
        shaping_policy.enabled = self.ts_enabled
        shaping_policy.averageBandwidth = self.ts_average_bandwidth * 1000
        shaping_policy.peakBandwidth = self.ts_peak_bandwidth * 1000
        shaping_policy.burstSize = self.ts_burst_size * 1024
        return shaping_policy

    def create_teaming_policy(self):
        """
        Create a NIC Teaming Policy
        Returns: NIC Teaming Policy object
        """
        # Only configure teaming policy if an option is defined
        if not all(option is None for option in [self.teaming_load_balancing,
                                                 self.teaming_failure_detection,
                                                 self.teaming_notify_switches,
                                                 self.teaming_failback,
                                                 self.teaming_failover_order_active,
                                                 self.teaming_failover_order_standby]):
            teaming_policy = vim.host.NetworkPolicy.NicTeamingPolicy()
            teaming_policy.policy = self.teaming_load_balancing
            # NOTE: 'teaming_inbound_policy' is deprecated and the following if statement should be removed in 2.11
            if self.teaming_inbound_policy:
                teaming_policy.reversePolicy = self.teaming_inbound_policy
            else:
                # Deprecated since VI API 5.1. The system default (true) will be used
                teaming_policy.reversePolicy = True
            teaming_policy.notifySwitches = self.teaming_notify_switches
            # NOTE: 'teaming_rolling_order' is deprecated and the following if statement should be removed in 2.11
            if self.teaming_rolling_order:
                teaming_policy.rollingOrder = self.teaming_rolling_order
            else:
                # this option is called 'failback' in the vSphere Client
                # rollingOrder also uses the opposite value displayed in the client
                if self.teaming_failback is None:
                    teaming_policy.rollingOrder = None
                else:
                    teaming_policy.rollingOrder = not self.teaming_failback
            if self.teaming_failover_order_active is None and self.teaming_failover_order_standby is None:
                teaming_policy.nicOrder = None
            else:
                teaming_policy.nicOrder = self.create_nic_order_policy()
            if self.teaming_failure_detection is None:
                teaming_policy.failureCriteria = None
            else:
                teaming_policy.failureCriteria = self.create_nic_failure_policy()
            return teaming_policy
        return None

    def create_nic_order_policy(self):
        """
        Create a NIC order Policy
        Returns: NIC order Policy object
        """
        for active_nic in self.teaming_failover_order_active:
            if active_nic not in self.switch_object.spec.bridge.nicDevice:
                self.module.fail_json(
                    msg="NIC '%s' (active) is not configured on vSwitch '%s'" % (active_nic, self.switch)
                )
        for standby_nic in self.teaming_failover_order_standby:
            if standby_nic not in self.switch_object.spec.bridge.nicDevice:
                self.module.fail_json(
                    msg="NIC '%s' (standby) is not configured on vSwitch '%s'" % (standby_nic, self.switch)
                )
        nic_order = vim.host.NetworkPolicy.NicOrderPolicy()
        nic_order.activeNic = self.teaming_failover_order_active
        nic_order.standbyNic = self.teaming_failover_order_standby
        return nic_order

    def create_nic_failure_policy(self):
        """
        Create a NIC Failure Criteria Policy
        Returns: NIC Failure Criteria Policy object
        """
        failure_criteria = vim.host.NetworkPolicy.NicFailureCriteria()
        if self.teaming_failure_detection == "link_status_only":
            failure_criteria.checkBeacon = False
        elif self.teaming_failure_detection == "beacon_probing":
            failure_criteria.checkBeacon = True
        elif self.teaming_failure_detection is None:
            failure_criteria = None
        # The following properties are deprecated since VI API 5.1. Default values are used
        failure_criteria.fullDuplex = False
        failure_criteria.percentage = 0
        failure_criteria.checkErrorPercent = False
        failure_criteria.checkDuplex = False
        failure_criteria.speed = 10
        failure_criteria.checkSpeed = 'minimum'
        return failure_criteria


def main():
    """Main"""
    argument_spec = vmware_argument_spec()
    argument_spec.update(dict(
        portgroup=dict(type='str', required=True, aliases=['portgroup_name']),
        switch=dict(type='str', required=True, aliases=['switch_name', 'vswitch']),
        vlan_id=dict(type='int', required=False, default=0, aliases=['vlan']),
        hosts=dict(type='list', aliases=['esxi_hostname']),
        cluster_name=dict(type='str', aliases=['cluster']),
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        security=dict(
            type='dict',
            options=dict(
                promiscuous_mode=dict(type='bool'),
                forged_transmits=dict(type='bool'),
                mac_changes=dict(type='bool'),
            ),
            aliases=['security_policy', 'network_policy']
        ),
        traffic_shaping=dict(
            type='dict',
            options=dict(
                enabled=dict(type='bool'),
                average_bandwidth=dict(type='int'),
                peak_bandwidth=dict(type='int'),
                burst_size=dict(type='int'),
            ),
        ),
        teaming=dict(
            type='dict',
            options=dict(
                load_balancing=dict(
                    type='str',
                    choices=[
                        None,
                        'loadbalance_ip',
                        'loadbalance_srcmac',
                        'loadbalance_srcid',
                        'failover_explicit',
                    ],
                    aliases=['load_balance_policy'],
                ),
                network_failure_detection=dict(
                    type='str',
                    choices=['link_status_only', 'beacon_probing']
                ),
                notify_switches=dict(type='bool'),
                failback=dict(type='bool'),
                active_adapters=dict(type='list'),
                standby_adapters=dict(type='list'),
                # NOTE: Deprecated from 2.11 onwards
                inbound_policy=dict(type='bool'),
                rolling_order=dict(type='bool'),
            ),
            aliases=['teaming_policy']
        ),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_one_of=[
            ['cluster_name', 'hosts'],
        ],
        supports_check_mode=True
    )

    try:
        host_portgroup = VMwareHostPortGroup(module)
        host_portgroup.process_state()
    except vmodl.RuntimeFault as runtime_fault:
        module.fail_json(msg=to_native(runtime_fault.msg))
    except vmodl.MethodFault as method_fault:
        module.fail_json(msg=to_native(method_fault.msg))
    except Exception as e:
        module.fail_json(msg=to_native(e))


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2018, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'certified'}

DOCUMENTATION = r'''
---
module: bigip_software_install
short_description: Install software images on a BIG-IP
description:
  - Install new images on a BIG-IP.
version_added: 2.7
options:
  image:
    description:
      - Image to install on the remote device.
    type: str
  volume:
    description:
      - The volume to install the software image to.
    type: str
  state:
    description:
      - When C(installed), ensures that the software is installed on the volume
        and the volume is set to be booted from. The device is B(not) rebooted
        into the new software.
      - When C(activated), performs the same operation as C(installed), but
        the system is rebooted to the new software.
    type: str
    choices:
      - activated
      - installed
    default: activated
extends_documentation_fragment: f5
author:
  - Tim Rupp (@caphrim007)
  - Wojciech Wypior (@wojtek0806)
'''
EXAMPLES = r'''
- name: Ensure an existing image is installed in specified volume
  bigip_software_install:
    image: BIGIP-13.0.0.0.0.1645.iso
    volume: HD1.2
    state: installed
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
  delegate_to: localhost

- name: Ensure an existing image is activated in specified volume
  bigip_software_install:
    image: BIGIP-13.0.0.0.0.1645.iso
    state: activated
    volume: HD1.2
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
  delegate_to: localhost
'''

RETURN = r'''
# only common fields returned
'''

import time
import ssl

from ansible.module_utils.six.moves.urllib.error import URLError
from ansible.module_utils.urls import urlparse
from ansible.module_utils.basic import AnsibleModule

try:
    from library.module_utils.network.f5.bigip import F5RestClient
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import AnsibleF5Parameters
    from library.module_utils.network.f5.common import f5_argument_spec
except ImportError:
    from ansible.module_utils.network.f5.bigip import F5RestClient
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import AnsibleF5Parameters
    from ansible.module_utils.network.f5.common import f5_argument_spec


class Parameters(AnsibleF5Parameters):
    api_map = {

    }

    api_attributes = [
        'options',
        'volume',
    ]

    returnables = [

    ]

    updatables = [

    ]


class ApiParameters(Parameters):
    @property
    def image_names(self):
        result = []
        result += self.read_image_from_device('image')
        result += self.read_image_from_device('hotfix')
        return result

    def read_image_from_device(self, t):
        uri = "https://{0}:{1}/mgmt/tm/sys/software/{2}".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
            t,
        )
        resp = self.client.api.get(uri)
        try:
            response = resp.json()
        except ValueError:
            return []

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                return []
            else:
                return []
        if 'items' not in response:
            return []
        return [x['name'].split('/')[0] for x in response['items']]


class ModuleParameters(Parameters):
    @property
    def version(self):
        if self._values['version']:
            return self._values['version']

        self._values['version'] = self.image_info['version']
        return self._values['version']

    @property
    def build(self):
        # Return cached copy if we have it
        if self._values['build']:
            return self._values['build']

        # Otherwise, get copy from image info cache
        self._values['build'] = self.image_info['build']
        return self._values['build']

    @property
    def image_info(self):
        if self._values['image_info']:
            image = self._values['image_info']
        else:
            # Otherwise, get a new copy and store in cache
            image = self.read_image()
            self._values['image_info'] = image
        return image

    @property
    def image_type(self):
        if self._values['image_type']:
            return self._values['image_type']
        if 'software:image' in self.image_info['kind']:
            self._values['image_type'] = 'image'
        else:
            self._values['image_type'] = 'hotfix'
        return self._values['image_type']

    def read_image(self):
        image = self.read_image_from_device(type='image')
        if image:
            return image
        image = self.read_image_from_device(type='hotfix')
        if image:
            return image
        return None

    def read_image_from_device(self, type):
        uri = "https://{0}:{1}/mgmt/tm/sys/software/{2}/".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
            type,
        )
        resp = self.client.api.get(uri)

        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'items' in response:
            for item in response['items']:
                if item['name'].startswith(self.image):
                    return item


class Changes(Parameters):
    def to_return(self):
        result = {}
        try:
            for returnable in self.returnables:
                result[returnable] = getattr(self, returnable)
            result = self._filter_params(result)
        except Exception:
            pass
        return result


class UsableChanges(Changes):
    pass


class ReportableChanges(Changes):
    pass


class Difference(object):
    def __init__(self, want, have=None):
        self.want = want
        self.have = have

    def compare(self, param):
        try:
            result = getattr(self, param)
            return result
        except AttributeError:
            return self.__default(param)

    def __default(self, param):
        attr1 = getattr(self.want, param)
        try:
            attr2 = getattr(self.have, param)
            if attr1 != attr2:
                return attr1
        except AttributeError:
            return attr1


class ModuleManager(object):
    def __init__(self, *args, **kwargs):
        self.module = kwargs.get('module', None)
        self.client = F5RestClient(**self.module.params)
        self.want = ModuleParameters(params=self.module.params, client=self.client)
        self.have = ApiParameters(client=self.client)
        self.changes = UsableChanges()
        self.volume_url = None

    def _set_changed_options(self):
        changed = {}
        for key in Parameters.returnables:
            if getattr(self.want, key) is not None:
                changed[key] = getattr(self.want, key)
        if changed:
            self.changes = UsableChanges(params=changed)

    def _update_changed_options(self):
        diff = Difference(self.want, self.have)
        updatables = Parameters.updatables
        changed = dict()
        for k in updatables:
            change = diff.compare(k)
            if change is None:
                continue
            else:
                if isinstance(change, dict):
                    changed.update(change)
                else:
                    changed[k] = change
        if changed:
            self.changes = UsableChanges(params=changed)
            return True
        return False

    def should_update(self):
        result = self._update_changed_options()
        if result:
            return True
        return False

    def exec_module(self):
        result = dict()

        changed = self.present()

        reportable = ReportableChanges(params=self.changes.to_return())
        changes = reportable.to_return()
        result.update(**changes)
        result.update(dict(changed=changed))
        self._announce_deprecations(result)
        return result

    def _announce_deprecations(self, result):
        warnings = result.pop('__warnings', [])
        for warning in warnings:
            self.client.module.deprecate(
                msg=warning['msg'],
                version=warning['version']
            )

    def present(self):
        if self.exists():
            return False
        else:
            return self.update()

    def _set_volume_url(self, item):
        path = urlparse(item['selfLink']).path
        self.volume_url = "https://{0}:{1}{2}".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
            path
        )

    def exists(self):
        uri = "https://{0}:{1}/mgmt/tm/sys/software/volume/".format(
            self.client.provider['server'],
            self.client.provider['server_port']
        )
        resp = self.client.api.get(uri)

        try:
            collection = resp.json()
        except ValueError:
            return False

        for item in collection['items']:
            if item['name'].startswith(self.want.volume):
                self._set_volume_url(item)
                break

        if not self.volume_url:
            self.volume_url = uri + self.want.volume

        resp = self.client.api.get(self.volume_url)

        try:
            response = resp.json()
        except ValueError:
            return False

        if resp.status == 404 or 'code' in response and response['code'] == 404:
            return False

        # version key can be missing in the event that an existing volume has
        # no installed software in it.
        if self.want.version != response.get('version', None):
            return False
        if self.want.build != response.get('build', None):
            return False

        if self.want.state == 'installed':
            return True
        if self.want.state == 'activated':
            if 'defaultBootLocation' in response['media'][0]:
                return True
        return False

    def volume_exists(self):
        resp = self.client.api.get(self.volume_url)

        try:
            response = resp.json()
        except ValueError:
            return False
        if resp.status == 404 or 'code' in response and response['code'] == 404:
            return False
        return True

    def update(self):
        if self.module.check_mode:
            return True

        if self.want.image and self.want.image not in self.have.image_names:
            raise F5ModuleError(
                "The specified image was not found on the device."
            )

        options = list()
        if not self.volume_exists():
            options.append({'create-volume': True})
        if self.want.state == 'activated':
            options.append({'reboot': True})
        self.want.update({'options': options})

        self.update_on_device()
        self.wait_for_software_install_on_device()
        if self.want.state == 'activated':
            self.wait_for_device_reboot()
        return True

    def update_on_device(self):
        params = {
            "command": "install",
            "name": self.want.image,
        }
        params.update(self.want.api_params())

        uri = "https://{0}:{1}/mgmt/tm/sys/software/{2}".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
            self.want.image_type
        )
        resp = self.client.api.post(uri, json=params)
        try:
            response = resp.json()
            if 'commandResult' in response and len(response['commandResult'].strip()) > 0:
                raise F5ModuleError(response['commandResult'])
        except ValueError as ex:
            raise F5ModuleError(str(ex))
        if 'code' in response and response['code'] in [400, 403]:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)
        return True

    def wait_for_device_reboot(self):
        while True:
            time.sleep(5)
            try:
                self.client.reconnect()
                volume = self.read_volume_from_device()
                if volume is None:
                    continue
                if 'active' in volume and volume['active'] is True:
                    break
            except F5ModuleError:
                # Handle all exceptions because if the system is offline (for a
                # reboot) the REST client will raise exceptions about
                # connections
                pass

    def wait_for_software_install_on_device(self):
        # We need to delay this slightly in case the the volume needs to be
        # created first
        for dummy in range(10):
            try:
                if self.volume_exists():
                    break
            except F5ModuleError:
                pass
            time.sleep(5)
        while True:
            time.sleep(10)
            volume = self.read_volume_from_device()
            if volume is None or 'status' not in volume:
                self.client.reconnect()
                continue
            if volume['status'] == 'complete':
                break
            elif volume['status'] == 'failed':
                raise F5ModuleError

    def read_volume_from_device(self):
        try:
            resp = self.client.api.get(self.volume_url)
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))
        except ssl.SSLError:
            # Suggests BIG-IP is still in the middle of restarting itself or
            # restjavad is restarting.
            return None
        except URLError:
            # At times during reboot BIG-IP will reset or timeout connections so we catch and pass this here.
            return None

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)
        return response


class ArgumentSpec(object):
    def __init__(self):
        self.supports_check_mode = True
        argument_spec = dict(
            image=dict(),
            volume=dict(),
            state=dict(
                default='activated',
                choices=['activated', 'installed']
            ),
        )
        self.argument_spec = {}
        self.argument_spec.update(f5_argument_spec)
        self.argument_spec.update(argument_spec)


def main():
    spec = ArgumentSpec()

    module = AnsibleModule(
        argument_spec=spec.argument_spec,
        supports_check_mode=spec.supports_check_mode,
    )

    try:
        mm = ModuleManager(module=module)
        results = mm.exec_module()
        module.exit_json(**results)
    except F5ModuleError as ex:
        module.fail_json(msg=str(ex))


if __name__ == '__main__':
    main()



# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.module_utils.six import string_types
from ansible.playbook.attribute import FieldAttribute
from ansible.utils.collection_loader import AnsibleCollectionLoader


def _ensure_default_collection(collection_list=None):
    default_collection = AnsibleCollectionLoader().default_collection

    if collection_list is None:
        collection_list = []

    if default_collection:  # FIXME: exclude role tasks?
        if isinstance(collection_list, string_types):
            collection_list = [collection_list]

        if default_collection not in collection_list:
            collection_list.insert(0, default_collection)

    # if there's something in the list, ensure that builtin or legacy is always there too
    if collection_list and 'ansible.builtin' not in collection_list and 'ansible.legacy' not in collection_list:
        collection_list.append('ansible.legacy')

    return collection_list


class CollectionSearch:

    # this needs to be populated before we can resolve tasks/roles/etc
    _collections = FieldAttribute(isa='list', listof=string_types, priority=100, default=_ensure_default_collection)

    def _load_collections(self, attr, ds):
        # this will only be called if someone specified a value; call the shared value
        _ensure_default_collection(collection_list=ds)

        if not ds:  # don't return an empty collection list, just return None
            return None

        return ds



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: win_domain_group
version_added: '2.4'
short_description: Creates, modifies or removes domain groups
description:
- Creates, modifies or removes groups in Active Directory.
- For local groups, use the M(win_group) module instead.
options:
  attributes:
    description:
    - A dict of custom LDAP attributes to set on the group.
    - This can be used to set custom attributes that are not exposed as module
      parameters, e.g. C(mail).
    - See the examples on how to format this parameter.
    type: dict
  category:
    description:
    - The category of the group, this is the value to assign to the LDAP
      C(groupType) attribute.
    - If a new group is created then C(security) will be used by default.
    type: str
    choices: [ distribution, security ]
  description:
    description:
    - The value to be assigned to the LDAP C(description) attribute.
    type: str
  display_name:
    description:
    - The value to assign to the LDAP C(displayName) attribute.
    type: str
  domain_username:
    description:
    - The username to use when interacting with AD.
    - If this is not set then the user Ansible used to log in with will be
      used instead.
    type: str
  domain_password:
    description:
    - The password for C(username).
    type: str
  domain_server:
    description:
    - Specifies the Active Directory Domain Services instance to connect to.
    - Can be in the form of an FQDN or NetBIOS name.
    - If not specified then the value is based on the domain of the computer
      running PowerShell.
    type: str
    version_added: '2.5'
  ignore_protection:
    description:
    - Will ignore the C(ProtectedFromAccidentalDeletion) flag when deleting or
      moving a group.
    - The module will fail if one of these actions need to occur and this value
      is set to C(no).
    type: bool
    default: no
  managed_by:
    description:
    - The value to be assigned to the LDAP C(managedBy) attribute.
    - This value can be in the forms C(Distinguished Name), C(objectGUID),
      C(objectSid) or C(sAMAccountName), see examples for more details.
    type: str
  name:
    description:
    - The name of the group to create, modify or remove.
    - This value can be in the forms C(Distinguished Name), C(objectGUID),
      C(objectSid) or C(sAMAccountName), see examples for more details.
    type: str
    required: yes
  organizational_unit:
    description:
    - The full LDAP path to create or move the group to.
    - This should be the path to the parent object to create or move the group to.
    - See examples for details of how this path is formed.
    type: str
    aliases: [ ou, path ]
  protect:
    description:
    - Will set the C(ProtectedFromAccidentalDeletion) flag based on this value.
    - This flag stops a user from deleting or moving a group to a different
      path.
    type: bool
  scope:
    description:
    - The scope of the group.
    - If C(state=present) and the group doesn't exist then this must be set.
    type: str
    choices: [domainlocal, global, universal]
  state:
    description:
    - If C(state=present) this module will ensure the group is created and is
      configured accordingly.
    - If C(state=absent) this module will delete the group if it exists
    type: str
    choices: [ absent, present ]
    default: present
notes:
- This must be run on a host that has the ActiveDirectory powershell module installed.
seealso:
- module: win_domain
- module: win_domain_controller
- module: win_domain_computer
- module: win_domain_membership
- module: win_domain_user
- module: win_group
- module: win_group_membership
author:
- Jordan Borean (@jborean93)
'''

EXAMPLES = r'''
- name: Ensure the group Cow exists using sAMAccountName
  win_domain_group:
    name: Cow
    scope: global
    path: OU=groups,DC=ansible,DC=local

- name: Ensure the group Cow doesn't exist using the Distinguished Name
  win_domain_group:
    name: CN=Cow,OU=groups,DC=ansible,DC=local
    state: absent

- name: Delete group ignoring the protection flag
  win_domain_group:
    name: Cow
    state: absent
    ignore_protection: yes

- name: Create group with delete protection enabled and custom attributes
  win_domain_group:
    name: Ansible Users
    scope: domainlocal
    category: security
    attributes:
      mail: helpdesk@ansible.com
      wWWHomePage: www.ansible.com
    ignore_protection: yes

- name: Change the OU of a group using the SID and ignore the protection flag
  win_domain_group:
    name: S-1-5-21-2171456218-3732823212-122182344-1189
    scope: global
    organizational_unit: OU=groups,DC=ansible,DC=local
    ignore_protection: yes

- name: Add managed_by user
  win_domain_group:
    name: Group Name Here
    managed_by: Domain Admins

- name: Add group and specify the AD domain services to use for the create
  win_domain_group:
    name: Test Group
    domain_username: user@CORP.ANSIBLE.COM
    domain_password: Password01!
    domain_server: corp-DC12.corp.ansible.com
    scope: domainlocal
'''

RETURN = r'''
attributes:
  description: Custom attributes that were set by the module. This does not
    show all the custom attributes rather just the ones that were set by the
    module.
  returned: group exists and attributes are set on the module invocation
  type: dict
  sample:
    mail: 'helpdesk@ansible.com'
    wWWHomePage: 'www.ansible.com'
canonical_name:
  description: The canonical name of the group.
  returned: group exists
  type: str
  sample: ansible.local/groups/Cow
category:
  description: The Group type value of the group, i.e. Security or Distribution.
  returned: group exists
  type: str
  sample: Security
description:
  description: The Description of the group.
  returned: group exists
  type: str
  sample: Group Description
display_name:
  description: The Display name of the group.
  returned: group exists
  type: str
  sample: Users who connect through RDP
distinguished_name:
  description: The full Distinguished Name of the group.
  returned: group exists
  type: str
  sample: CN=Cow,OU=groups,DC=ansible,DC=local
group_scope:
  description: The Group scope value of the group.
  returned: group exists
  type: str
  sample: Universal
guid:
  description: The guid of the group.
  returned: group exists
  type: str
  sample: 512a9adb-3fc0-4a26-9df0-e6ea1740cf45
managed_by:
  description: The full Distinguished Name of the AD object that is set on the
    managedBy attribute.
  returned: group exists
  type: str
  sample: CN=Domain Admins,CN=Users,DC=ansible,DC=local
name:
  description: The name of the group.
  returned: group exists
  type: str
  sample: Cow
protected_from_accidental_deletion:
  description: Whether the group is protected from accidental deletion.
  returned: group exists
  type: bool
  sample: true
sid:
  description: The Security ID of the group.
  returned: group exists
  type: str
  sample: S-1-5-21-2171456218-3732823212-122182344-1189
created:
  description: Whether a group was created
  returned: always
  type: bool
  sample: true
'''



#!/usr/bin/python
# coding: utf-8 -*-

# (c) 2017, John Westcott IV <john.westcott.iv@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: tower_receive
author: "John Westcott IV (@john-westcott-iv)"
version_added: "2.8"
short_description: Receive assets from Ansible Tower.
description:
    - Receive assets from Ansible Tower. See
      U(https://www.ansible.com/tower) for an overview.
options:
    all:
      description:
        - Export all assets
      type: bool
      default: 'False'
    organization:
      description:
        - List of organization names to export
      default: []
    user:
      description:
        - List of user names to export
      default: []
    team:
      description:
        - List of team names to export
      default: []
    credential_type:
      description:
        - List of credential type names to export
      default: []
    credential:
      description:
        - List of credential names to export
      default: []
    notification_template:
      description:
        - List of notification template names to export
      default: []
    inventory_script:
      description:
        - List of inventory script names to export
      default: []
    inventory:
      description:
        - List of inventory names to export
      default: []
    project:
      description:
        - List of project names to export
      default: []
    job_template:
      description:
        - List of job template names to export
      default: []
    workflow:
      description:
        - List of workflow names to export
      default: []

requirements:
  - "ansible-tower-cli >= 3.3.0"

notes:
  - Specifying a name of "all" for any asset type will export all items of that asset type.

extends_documentation_fragment: tower
'''

EXAMPLES = '''
- name: Export all tower assets
  tower_receive:
    all: True
    tower_config_file: "~/tower_cli.cfg"

- name: Export all inventories
  tower_receive:
    inventory:
      - all

- name: Export a job template named "My Template" and all Credentials
  tower_receive:
    job_template:
      - "My Template"
    credential:
      - all
'''

RETURN = '''
assets:
    description: The exported assets
    returned: success
    type: dict
    sample: [ {}, {} ]
'''

from ansible.module_utils.ansible_tower import TowerModule, tower_auth_config, HAS_TOWER_CLI

try:
    from tower_cli.cli.transfer.receive import Receiver
    from tower_cli.cli.transfer.common import SEND_ORDER
    from tower_cli.utils.exceptions import TowerCLIError

    from tower_cli.conf import settings
    TOWER_CLI_HAS_EXPORT = True
except ImportError:
    TOWER_CLI_HAS_EXPORT = False


def main():
    argument_spec = dict(
        all=dict(type='bool', default=False),
        credential=dict(type='list', default=[]),
        credential_type=dict(type='list', default=[]),
        inventory=dict(type='list', default=[]),
        inventory_script=dict(type='list', default=[]),
        job_template=dict(type='list', default=[]),
        notification_template=dict(type='list', default=[]),
        organization=dict(type='list', default=[]),
        project=dict(type='list', default=[]),
        team=dict(type='list', default=[]),
        user=dict(type='list', default=[]),
        workflow=dict(type='list', default=[]),
    )

    module = TowerModule(argument_spec=argument_spec, supports_check_mode=False)

    if not HAS_TOWER_CLI:
        module.fail_json(msg='ansible-tower-cli required for this module')

    if not TOWER_CLI_HAS_EXPORT:
        module.fail_json(msg='ansible-tower-cli version does not support export')

    export_all = module.params.get('all')
    assets_to_export = {}
    for asset_type in SEND_ORDER:
        assets_to_export[asset_type] = module.params.get(asset_type)

    result = dict(
        assets=None,
        changed=False,
        message='',
    )

    tower_auth = tower_auth_config(module)
    with settings.runtime_values(**tower_auth):
        try:
            receiver = Receiver()
            result['assets'] = receiver.export_assets(all=export_all, asset_input=assets_to_export)
            module.exit_json(**result)
        except TowerCLIError as e:
            result['message'] = e.message
            module.fail_json(msg='Receive Failed', **result)


if __name__ == '__main__':
    main()



# Copyright (c) 2017 Cisco and/or its affiliates.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

import json

from units.compat.mock import patch
from units.compat import unittest
from ansible.module_utils.network.nso import nso


MODULE_PREFIX_MAP = '''
{
  "ansible-nso": "an",
  "test": "test",
  "tailf-ncs": "ncs"
}
'''


SCHEMA_DATA = {
    '/an:id-name-leaf': '''
{
  "meta": {
    "prefix": "an",
    "namespace": "http://github.com/ansible/nso",
    "types": {
      "http://github.com/ansible/nso:id-name-t": [
        {
          "name": "http://github.com/ansible/nso:id-name-t",
          "enumeration": [
            {
              "label": "id-one"
            },
            {
              "label": "id-two"
            }
          ]
        },
        {
          "name": "identityref"
        }
      ]
    },
    "keypath": "/an:id-name-leaf"
  },
  "data": {
    "kind": "leaf",
    "type": {
      "namespace": "http://github.com/ansible/nso",
      "name": "id-name-t"
    },
    "name": "id-name-leaf",
    "qname": "an:id-name-leaf"
  }
}''',
    '/an:id-name-values': '''
{
  "meta": {
    "prefix": "an",
    "namespace": "http://github.com/ansible/nso",
    "types": {},
    "keypath": "/an:id-name-values"
  },
  "data": {
    "kind": "container",
    "name": "id-name-values",
    "qname": "an:id-name-values",
    "children": [
      {
        "kind": "list",
        "name": "id-name-value",
        "qname": "an:id-name-value",
        "key": [
          "name"
        ]
      }
    ]
  }
}
''',
    '/an:id-name-values/id-name-value': '''
{
  "meta": {
    "prefix": "an",
    "namespace": "http://github.com/ansible/nso",
    "types": {
      "http://github.com/ansible/nso:id-name-t": [
        {
          "name": "http://github.com/ansible/nso:id-name-t",
          "enumeration": [
            {
              "label": "id-one"
            },
            {
              "label": "id-two"
            }
          ]
        },
        {
          "name": "identityref"
        }
      ]
    },
    "keypath": "/an:id-name-values/id-name-value"
  },
  "data": {
    "kind": "list",
    "name": "id-name-value",
    "qname": "an:id-name-value",
    "key": [
      "name"
    ],
    "children": [
      {
        "kind": "key",
        "name": "name",
        "qname": "an:name",
        "type": {
          "namespace": "http://github.com/ansible/nso",
          "name": "id-name-t"
        }
      },
      {
        "kind": "leaf",
        "type": {
          "primitive": true,
          "name": "string"
        },
        "name": "value",
        "qname": "an:value"
      }
    ]
  }
}
''',
    '/test:test': '''
{
    "meta": {
        "types": {
            "http://example.com/test:t15": [
               {
                  "leaf_type":[
                     {
                        "name":"string"
                     }
                  ],
                  "list_type":[
                     {
                        "name":"http://example.com/test:t15",
                        "leaf-list":true
                     }
                  ]
               }
            ]
        }
    },
    "data": {
        "kind": "list",
        "name":"test",
        "qname":"test:test",
        "key":["name"],
        "children": [
            {
                "kind": "key",
                "name": "name",
                "qname": "test:name",
                "type": {"name":"string","primitive":true}
            },
            {
                "kind": "choice",
                "name": "test-choice",
                "qname": "test:test-choice",
                "cases": [
                    {
                        "kind": "case",
                        "name": "direct-child-case",
                        "qname":"test:direct-child-case",
                        "children":[
                            {
                                "kind": "leaf",
                                "name": "direct-child",
                                "qname": "test:direct-child",
                                "type": {"name":"string","primitive":true}
                            }
                        ]
                    },
                    {
                        "kind":"case","name":"nested-child-case","qname":"test:nested-child-case",
                        "children": [
                            {
                                "kind": "choice",
                                "name": "nested-choice",
                                "qname": "test:nested-choice",
                                "cases": [
                                    {
                                        "kind":"case","name":"nested-child","qname":"test:nested-child",
                                        "children": [
                                            {
                                               "kind": "leaf",
                                               "name":"nested-child",
                                                "qname":"test:nested-child",
                                               "type":{"name":"string","primitive":true}}
                                         ]
                                    }
                             ]
                            }
                        ]
                    }
                ]
            },
            {
               "kind":"leaf-list",
               "name":"device-list",
               "qname":"test:device-list",
               "type": {
                  "namespace":"http://example.com/test",
                  "name":"t15"
               }
            }
        ]
    }
}
''',
    '/test:test/device-list': '''
{
    "meta": {
        "types": {
            "http://example.com/test:t15": [
               {
                  "leaf_type":[
                     {
                        "name":"string"
                     }
                  ],
                  "list_type":[
                     {
                        "name":"http://example.com/test:t15",
                        "leaf-list":true
                     }
                  ]
               }
            ]
        }
    },
    "data": {
        "kind":"leaf-list",
        "name":"device-list",
        "qname":"test:device-list",
        "type": {
           "namespace":"http://example.com/test",
           "name":"t15"
        }
    }
}
''',
    '/test:deps': '''
{
    "meta": {
    },
    "data": {
        "kind":"container",
        "name":"deps",
        "qname":"test:deps",
        "children": [
            {
                "kind": "leaf",
                "type": {
                  "primitive": true,
                  "name": "string"
                },
                "name": "a",
                "qname": "test:a",
                "deps": ["/test:deps/c"]
            },
            {
                "kind": "leaf",
                "type": {
                  "primitive": true,
                  "name": "string"
                },
                "name": "b",
                "qname": "test:b",
                "deps": ["/test:deps/a"]
            },
            {
                "kind": "leaf",
                "type": {
                  "primitive": true,
                  "name": "string"
                },
                "name": "c",
                "qname": "test:c"
            }
        ]
    }
}
'''
}


class MockResponse(object):
    def __init__(self, method, params, code, body, headers=None):
        if headers is None:
            headers = {}

        self.method = method
        self.params = params

        self.code = code
        self.body = body
        self.headers = dict(headers)

    def read(self):
        return self.body


def mock_call(calls, url, timeout, validate_certs, data=None, headers=None, method=None):
    result = calls[0]
    del calls[0]

    request = json.loads(data)
    if result.method != request['method']:
        raise ValueError('expected method {0}({1}), got {2}({3})'.format(
            result.method, result.params,
            request['method'], request['params']))

    for key, value in result.params.items():
        if key not in request['params']:
            raise ValueError('{0} not in parameters'.format(key))
        if value != request['params'][key]:
            raise ValueError('expected {0} to be {1}, got {2}'.format(
                key, value, request['params'][key]))

    return result


def get_schema_response(path):
    return MockResponse(
        'get_schema', {'path': path}, 200, '{{"result": {0}}}'.format(
            SCHEMA_DATA[path]))


class TestJsonRpc(unittest.TestCase):
    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_exists(self, open_url_mock):
        calls = [
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            MockResponse('exists', {'path': '/exists'}, 200, '{"result": {"exists": true}}'),
            MockResponse('exists', {'path': '/not-exists'}, 200, '{"result": {"exists": false}}')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)
        client = nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False)
        self.assertEquals(True, client.exists('/exists'))
        self.assertEquals(False, client.exists('/not-exists'))

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_exists_data_not_found(self, open_url_mock):
        calls = [
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            MockResponse('exists', {'path': '/list{missing-parent}/list{child}'}, 200, '{"error":{"type":"data.not_found"}}')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)
        client = nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False)
        self.assertEquals(False, client.exists('/list{missing-parent}/list{child}'))

        self.assertEqual(0, len(calls))


class TestValueBuilder(unittest.TestCase):
    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_identityref_leaf(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/an:id-name-leaf'),
            MockResponse('get_module_prefix_map', {}, 200, '{{"result": {0}}}'.format(MODULE_PREFIX_MAP))
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/an:id-name-leaf"
        schema_data = json.loads(
            SCHEMA_DATA['/an:id-name-leaf'])
        schema = schema_data['data']

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, 'ansible-nso:id-two', schema)
        values = list(vb.values)
        self.assertEquals(1, len(values))
        value = values[0]
        self.assertEquals(parent, value.path)
        self.assertEquals('set', value.state)
        self.assertEquals('an:id-two', value.value)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_identityref_key(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/an:id-name-values/id-name-value'),
            MockResponse('get_module_prefix_map', {}, 200, '{{"result": {0}}}'.format(MODULE_PREFIX_MAP)),
            MockResponse('exists', {'path': '/an:id-name-values/id-name-value{an:id-one}'}, 200, '{"result": {"exists": true}}')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/an:id-name-values"
        schema_data = json.loads(
            SCHEMA_DATA['/an:id-name-values/id-name-value'])
        schema = schema_data['data']

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, 'id-name-value', [{'name': 'ansible-nso:id-one', 'value': '1'}], schema)
        values = list(vb.values)
        self.assertEquals(1, len(values))
        value = values[0]
        self.assertEquals('{0}/id-name-value{{an:id-one}}/value'.format(parent), value.path)
        self.assertEquals('set', value.state)
        self.assertEquals('1', value.value)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_nested_choice(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/test:test'),
            MockResponse('exists', {'path': '/test:test{direct}'}, 200, '{"result": {"exists": true}}'),
            MockResponse('exists', {'path': '/test:test{nested}'}, 200, '{"result": {"exists": true}}')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/test:test"
        schema_data = json.loads(
            SCHEMA_DATA['/test:test'])
        schema = schema_data['data']

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, [{'name': 'direct', 'direct-child': 'direct-value'},
                                {'name': 'nested', 'nested-child': 'nested-value'}], schema)
        values = list(vb.values)
        self.assertEquals(2, len(values))
        value = values[0]
        self.assertEquals('{0}{{direct}}/direct-child'.format(parent), value.path)
        self.assertEquals('set', value.state)
        self.assertEquals('direct-value', value.value)

        value = values[1]
        self.assertEquals('{0}{{nested}}/nested-child'.format(parent), value.path)
        self.assertEquals('set', value.state)
        self.assertEquals('nested-value', value.value)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_leaf_list_type(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.4"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/test:test')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/test:test"
        schema_data = json.loads(
            SCHEMA_DATA['/test:test'])
        schema = schema_data['data']

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, {'device-list': ['one', 'two']}, schema)
        values = list(vb.values)
        self.assertEquals(1, len(values))
        value = values[0]
        self.assertEquals('{0}/device-list'.format(parent), value.path)
        self.assertEquals(['one', 'two'], value.value)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_leaf_list_type_45(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/test:test/device-list')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/test:test"
        schema_data = json.loads(
            SCHEMA_DATA['/test:test'])
        schema = schema_data['data']

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, {'device-list': ['one', 'two']}, schema)
        values = list(vb.values)
        self.assertEquals(3, len(values))
        value = values[0]
        self.assertEquals('{0}/device-list'.format(parent), value.path)
        self.assertEquals(nso.State.ABSENT, value.state)
        value = values[1]
        self.assertEquals('{0}/device-list{{one}}'.format(parent), value.path)
        self.assertEquals(nso.State.PRESENT, value.state)
        value = values[2]
        self.assertEquals('{0}/device-list{{two}}'.format(parent), value.path)
        self.assertEquals(nso.State.PRESENT, value.state)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_sort_by_deps(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/test:deps')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/test:deps"
        schema_data = json.loads(
            SCHEMA_DATA['/test:deps'])
        schema = schema_data['data']

        values = {
            'a': '1',
            'b': '2',
            'c': '3',
        }

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, values, schema)
        values = list(vb.values)
        self.assertEquals(3, len(values))
        value = values[0]
        self.assertEquals('{0}/c'.format(parent), value.path)
        self.assertEquals('3', value.value)
        value = values[1]
        self.assertEquals('{0}/a'.format(parent), value.path)
        self.assertEquals('1', value.value)
        value = values[2]
        self.assertEquals('{0}/b'.format(parent), value.path)
        self.assertEquals('2', value.value)

        self.assertEqual(0, len(calls))

    @patch('ansible.module_utils.network.nso.nso.open_url')
    def test_sort_by_deps_not_included(self, open_url_mock):
        calls = [
            MockResponse('get_system_setting', {'operation': 'version'}, 200, '{"result": "4.5"}'),
            MockResponse('new_trans', {}, 200, '{"result": {"th": 1}}'),
            get_schema_response('/test:deps')
        ]
        open_url_mock.side_effect = lambda *args, **kwargs: mock_call(calls, *args, **kwargs)

        parent = "/test:deps"
        schema_data = json.loads(
            SCHEMA_DATA['/test:deps'])
        schema = schema_data['data']

        values = {
            'a': '1',
            'b': '2'
        }

        vb = nso.ValueBuilder(nso.JsonRpc('http://localhost:8080/jsonrpc', 10, False))
        vb.build(parent, None, values, schema)
        values = list(vb.values)
        self.assertEquals(2, len(values))
        value = values[0]
        self.assertEquals('{0}/a'.format(parent), value.path)
        self.assertEquals('1', value.value)
        value = values[1]
        self.assertEquals('{0}/b'.format(parent), value.path)
        self.assertEquals('2', value.value)

        self.assertEqual(0, len(calls))


class TestVerifyVersion(unittest.TestCase):
    def test_valid_versions(self):
        self.assertTrue(nso.verify_version_str('5.0', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('5.1.1', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('5.1.1.2', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.6', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.6.2', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.6.2.1', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.5.1', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.5.2', [(4, 6), (4, 5, 1)]))
        self.assertTrue(nso.verify_version_str('4.5.1.2', [(4, 6), (4, 5, 1)]))

    def test_invalid_versions(self):
        self.assertFalse(nso.verify_version_str('4.4', [(4, 6), (4, 5, 1)]))
        self.assertFalse(nso.verify_version_str('4.4.1', [(4, 6), (4, 5, 1)]))
        self.assertFalse(nso.verify_version_str('4.4.1.2', [(4, 6), (4, 5, 1)]))
        self.assertFalse(nso.verify_version_str('4.5.0', [(4, 6), (4, 5, 1)]))


class TestValueSort(unittest.TestCase):
    def test_sort_parent_depend(self):
        values = [
            nso.ValueBuilder.Value('/test/list{entry}', '/test/list', 'CREATE', ['']),
            nso.ValueBuilder.Value('/test/list{entry}/description', '/test/list/description', 'TEST', ['']),
            nso.ValueBuilder.Value('/test/entry', '/test/entry', 'VALUE', ['/test/list', '/test/list/name'])
        ]

        result = [v.path for v in nso.ValueBuilder.sort_values(values)]

        self.assertEquals(['/test/list{entry}', '/test/entry', '/test/list{entry}/description'], result)

    def test_sort_break_direct_cycle(self):
        values = [
            nso.ValueBuilder.Value('/test/a', '/test/a', 'VALUE', ['/test/c']),
            nso.ValueBuilder.Value('/test/b', '/test/b', 'VALUE', ['/test/a']),
            nso.ValueBuilder.Value('/test/c', '/test/c', 'VALUE', ['/test/a'])
        ]

        result = [v.path for v in nso.ValueBuilder.sort_values(values)]

        self.assertEquals(['/test/a', '/test/b', '/test/c'], result)

    def test_sort_break_indirect_cycle(self):
        values = [
            nso.ValueBuilder.Value('/test/c', '/test/c', 'VALUE', ['/test/a']),
            nso.ValueBuilder.Value('/test/a', '/test/a', 'VALUE', ['/test/b']),
            nso.ValueBuilder.Value('/test/b', '/test/b', 'VALUE', ['/test/c'])
        ]

        result = [v.path for v in nso.ValueBuilder.sort_values(values)]

        self.assertEquals(['/test/a', '/test/c', '/test/b'], result)

    def test_sort_depend_on_self(self):
        values = [
            nso.ValueBuilder.Value('/test/a', '/test/a', 'VALUE', ['/test/a']),
            nso.ValueBuilder.Value('/test/b', '/test/b', 'VALUE', [])
        ]

        result = [v.path for v in nso.ValueBuilder.sort_values(values)]

        self.assertEqual(['/test/a', '/test/b'], result)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Cisco and/or its affiliates.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'certified'
}

DOCUMENTATION = '''
---
module: nso_show
extends_documentation_fragment: nso
short_description: Displays data from Cisco NSO.
description:
  - This module provides support for displaying data from Cisco NSO.
requirements:
  - Cisco NSO version 3.4.12 or higher, 4.1.9 or higher, 4.2.6 or higher,
    4.3.7 or higher, 4.4.5 or higher, 4.5 or higher.
author: "Claes Nästén (@cnasten)"
options:
  path:
    description: Path to NSO data.
    required: true
  operational:
    description: >
      Controls whether or not operational data is included in the result.
    type: bool
    default: false
version_added: "2.5"
'''

EXAMPLES = '''
- name: Show devices including operational data
  nso_show:
    url: http://localhost:8080/jsonrpc
    username: username
    password: password
    path: /ncs:devices/device
    operational: true
'''

RETURN = '''
output:
  description: Configuration
  returned: success
  type: dict
'''

from ansible.module_utils.network.nso.nso import connect, verify_version, nso_argument_spec
from ansible.module_utils.network.nso.nso import ModuleFailException, NsoException
from ansible.module_utils.basic import AnsibleModule


class NsoShow(object):
    REQUIRED_VERSIONS = [
        (4, 5),
        (4, 4, 5),
        (4, 3, 7),
        (4, 2, 6),
        (4, 1, 9),
        (3, 4, 12)
    ]

    def __init__(self, check_mode, client, path, operational):
        self._check_mode = check_mode
        self._client = client
        self._path = path
        self._operational = operational

    def main(self):
        if self._check_mode:
            return {}
        else:
            return self._client.show_config(self._path, self._operational)


def main():
    argument_spec = dict(
        path=dict(required=True, type='str'),
        operational=dict(required=False, type='bool', default=False)
    )
    argument_spec.update(nso_argument_spec)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True
    )
    p = module.params

    client = connect(p)
    nso_show = NsoShow(
        module.check_mode, client,
        p['path'], p['operational'])
    try:
        verify_version(client, NsoShow.REQUIRED_VERSIONS)

        output = nso_show.main()
        client.logout()
        module.exit_json(changed=False, output=output)
    except NsoException as ex:
        client.logout()
        module.fail_json(msg=ex.message)
    except ModuleFailException as ex:
        client.logout()
        module.fail_json(msg=ex.message)


if __name__ == '__main__':
    main()
